from mlapp.handlers.databases.mssql_handler import MssqlHandler
from mlapp.handlers.databases.mysql_handler import MySQLHandler
from mlapp.handlers.databases.postgres_handler import PostgresHandler
from mlapp.handlers.databases.snowflake_handler import SnowflakeHandler

try:
    from mlapp.handlers.file_storages.minio_storage_handler import MinioStorageHandler
except Exception as e:
    MinioStorageHandler = None

try:
    from mlapp.handlers.message_queues.rabbitmq_handler import RabbitMQHandler
except Exception as e:
    RabbitMQHandler = None

try:
    from mlapp.handlers.file_storages.azure_blob_handler import AzureBlobHandler
except Exception as e:
    AzureBlobHandler = None

try:
    from mlapp.handlers.message_queues.azure_service_bus_handler import AzureServicesBusHandler
except Exception as e:
    AzureServicesBusHandler = None

try:
    from mlapp.handlers.spark.spark_handler import SparkHandler
except Exception as e:
    SparkHandler = None

try:
    from handlers.hive_handler import HiveHandler
except Exception as e:
    HiveHandler = None

try:
    from mlapp.handlers.spark.livy_handler import LivyHandler
except Exception as e:
    LivyHandler = None

try:
    from mlapp.handlers.file_storages.boto_storage_handler import BotoStorageHandler
except Exception as e:
    BotoStorageHandler = None

try:
    from mlapp.handlers.file_storages.ibm_boto3_storage_handler import IBMBoto3StorageHandler
except Exception as e:
    IBMBoto3StorageHandler = None

try:
    from mlapp.handlers.message_queues.kafka_handler import KafkaHandler
except Exception as e:
    KafkaHandler = None

try:
    from mlapp.handlers.message_queues.kafka_kerberos_handler import KafkaKerberosHandler
except Exception as e:
    KafkaKerberosHandler = None

try:
    from mlapp.handlers.file_storages.aml_model_storage_handler import AmlModelStorageHandler
except Exception as e:
    AmlModelStorageHandler = None

try:
    from mlapp.handlers.file_storages.aml_run_storage_handler import AmlRunStorageHandler
except Exception as e:
    AmlRunStorageHandler = None

try:
    from mlapp.handlers.message_queues.aml_queue import AMLQueue
except Exception as e:
    AMLQueue = None

from mlapp.env_loader import EMPTY_STRING, ZERO


def environment_services(env=None):
    """
    Holds the mapping between the environment file and the input required for each handler
    :param env: Env class - from environs library
    """
    return {
        # default mysql db settings for Kubernetes deployment. See ML App Wiki documentation
        "mysql": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_DB', default=False),
            "handler": MySQLHandler,
            "type": "database",
            "settings": {
                "database_name": env.str(x.upper() + '_MYSQL_DATABASE_NAME', default=EMPTY_STRING),
                "hostname": env.str(x.upper() + '_MYSQL_HOSTNAME', default=EMPTY_STRING),
                "password": env.str(x.upper() + '_MYSQL_PASSWORD', default=EMPTY_STRING),
                "port": env.int(x.upper() + '_MYSQL_PORT', default=ZERO),
                "user_id": env.str(x.upper() + '_MYSQL_USER_ID', default=EMPTY_STRING),
                "options": {}
            }
        },
        # default minio file storage settings for Kubernetes deployment. See ML App Wiki documentation
        "minio": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_FS', default=False),
            "handler": MinioStorageHandler,
            "type": "file_storage",
            "settings": {
                'endPoint': env.str(x.upper() + '_MINIO_ENDPOINT', default=EMPTY_STRING),
                'accessKey': env.str(x.upper() + '_MINIO_ACCESS_KEY', default=EMPTY_STRING),
                'secretKey': env.str(x.upper() + '_MINIO_SECRET_KEY', default=EMPTY_STRING),
                'port': env.str(x.upper() + '_MINIO_PORT', default=EMPTY_STRING),
                'secure': env.bool(x.upper() + '_MINIO_SECURE', default=False),
                'region': env.str(x.upper() + '_MINIO_REGION', default=EMPTY_STRING)
            }
        },
        "azure_blob": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_FS', default=False),
            "type": "file_storage",
            "handler": AzureBlobHandler,
            "settings": {
                'accountName': env.str(x.upper() + '_AZURE_BLOB_ACCOUNT_NAME', default=EMPTY_STRING),
                'accountKey': env.str(x.upper() + '_AZURE_BLOB_ACCOUNT_KEY', default=EMPTY_STRING),
            }
        },
        "azureml_model_storage": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_FS', default=False),
            "type": "file_storage",
            "handler": AmlModelStorageHandler,
            "settings": {
                "local_storage_path": settings.get('local_storage_path'),
                "temporary_storage_path": settings.get('temporary_storage_path'),
                "file_store_buckets": settings.get('file_store_buckets')
            }
        },
        "azureml_run_storage": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_FS', default=False),
            "type": "file_storage",
            "handler": AmlRunStorageHandler,
            "settings": {
                "local_storage_path": settings.get('local_storage_path'),
                "temporary_storage_path": settings.get('temporary_storage_path'),
                "file_store_buckets": settings.get('file_store_buckets'),
                "database_name": env.str(x.upper() + '_AML_RUN_STORAGE_DATABASE_NAME', default=EMPTY_STRING),
                "hostname": env.str(x.upper() + '_AML_RUN_STORAGE_HOSTNAME', default=EMPTY_STRING),
                "password": env.str(x.upper() + '_AML_RUN_STORAGE_PASSWORD', default=EMPTY_STRING),
                "port": env.int(x.upper() + '_AML_RUN_STORAGE_PORT', default=ZERO),
                "user_id": env.str(x.upper() + '_AML_RUN_STORAGE_USER_ID', default=EMPTY_STRING),
                "options": {}
            }
        },
        "azure_service_bus": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_MQ', default=False),
            "type": "message_queue",
            "handler": AzureServicesBusHandler,
            "settings": {
                'hostname': env.str(x.upper() + '_AZURE_SERVICE_BUS_HOSTNAME', default=EMPTY_STRING),
                'shared_access_key_name':
                    env.str(x.upper() + '_AZURE_SERVICE_BUS_SHARED_ACCESS_KEY_NAME', default=EMPTY_STRING),
                'shared_access_key':
                    env.str(x.upper() + '_AZURE_SERVICE_BUS_SHARED_ACCESS_KEY', default=EMPTY_STRING),
                'entity_path': env.str(x.upper() + '_AZURE_SERVICE_BUS_ENTITY_PATH', default=EMPTY_STRING),
            }
        },
        "mssql": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_DB', default=False),
            "type": "database",
            "handler": MssqlHandler,
            "settings": {
                "database_name": env.str(x.upper() + '_MSSQL_DATABASE_NAME', default=EMPTY_STRING),
                "hostname": env.str(x.upper() + '_MSSQL_HOSTNAME', default=EMPTY_STRING),
                "password": env.str(x.upper() + '_MSSQL_PASSWORD', default=EMPTY_STRING),
                "port": env.int(x.upper() + '_MSSQL_PORT', default=ZERO),
                "user_id": env.str(x.upper() + '_MSSQL_USER_ID', default=EMPTY_STRING),
                "options": {}
            }
        },
        "snowflake": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_DB', default=False),
            "type": "database",
            "handler": SnowflakeHandler,
            "settings": {
                "account": env.str(x.upper() + '_SNOWFLAKE_ACCOUNT', default=EMPTY_STRING),
                "user": env.str(x.upper() + '_SNOWFLAKE_USER', default=EMPTY_STRING),
                "password": env.str(x.upper() + '_SNOWFLAKE_PASSWORD', default=EMPTY_STRING),
                "database": env.str(x.upper() + '_SNOWFLAKE_DATABASE', default=EMPTY_STRING),
                "schema": env.str(x.upper() + '_SNOWFLAKE_SCHEMA', default=EMPTY_STRING),
                "warehouse": env.str(x.upper() + '_SNOWFLAKE_WAREHOUSE', default=EMPTY_STRING),
                "role": env.str(x.upper() + '_SNOWFLAKE_ROLE', default=EMPTY_STRING),
                "options": {}
            }
        },
        "postgres": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_DB', default=False),
            "type": "database",
            "handler": PostgresHandler,
            "settings": {
                "database_name": env.str(x.upper() + '_POSTGRES_DATABASE_NAME', default=EMPTY_STRING),
                "hostname": env.str(x.upper() + '_POSTGRES_HOSTNAME', default=EMPTY_STRING),
                "password": env.str(x.upper() + '_POSTGRES_PASSWORD', default=EMPTY_STRING),
                "port": env.int(x.upper() + '_POSTGRES_PORT', default=ZERO),
                "user_id": env.str(x.upper() + '_POSTGRES_USER_ID', default=EMPTY_STRING),
                "ssl": env.bool(x.upper() + '_POSTGRES_SSL', default=False),
                "use_aws_IAM": env.bool(x.upper() + '_POSTGRES_USE_IAM', default=False),
                "aws_access_key": env.str(x.upper() + '_POSTGRES_IAM_ACCESS_KEY', default=EMPTY_STRING),
                "aws_secret_key": env.str(x.upper() + '_POSTGRES_IAM_SECRET_KEY', default=EMPTY_STRING),
                "options": {}
            }
        },
        "databricks": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_SPARK', default=False),
            "type": "spark",
            "handler": SparkHandler,
            "settings": {
                "spark.databricks.service.address": env.str(x.upper() + '_DATABRICKS_ADDRESS', default=EMPTY_STRING),
                "spark.databricks.service.token": env.str(x.upper() + '_DATABRICKS_TOKEN', default=EMPTY_STRING),
                "spark.databricks.service.clusterId": env.str(x.upper() + '_DATABRICKS_CLUSTER_ID', default=EMPTY_STRING),
                "spark.databricks.service.port": env.int(x.upper() + '_DATABRICKS_PORT', default=ZERO),
                "spark.databricks.service.orgId": env.int(x.upper() + '_DATABRICKS_ORGANIZATION_ID', default=ZERO),
                "enable_hive": env.str(x.upper() + '_ENABLE_HIVE', default=False),
            }
        },
        "spark": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_SPARK', default=False),
            "type": "spark",
            "handler": SparkHandler,
            "settings": {
                # cluster
                "spark.sql.warehouse.dir":  env.str(x.upper() + '_SPARK_SQL_WAREHOUSE_DIR', default=EMPTY_STRING),
                "hive.metastore.warehouse.dir":  env.str(x.upper() + '_SPARK_HIVE_METASTORE_WAREHOUSE_DIR', default=EMPTY_STRING),
                "hive.metastore.uris": env.str(x.upper() + '_SPARK_HIVE_METASTORE_URIS', default=EMPTY_STRING),
                "enable_hive": env.bool(x.upper() + '_SPARK_ENABLE_HIVE', default=False),
                "spark.driver.host":  env.str(x.upper() + '_SPARK_DRIVER_HOST', default=EMPTY_STRING),
                "spark.app.id":  env.str(x.upper() + '_SPARK_APP_ID', default=EMPTY_STRING),
                "spark.sql.catalogImplementation":  env.str(x.upper() + '_SPARK_SQL_CATALOG_IMPLEMENTATION', default=EMPTY_STRING),
                "spark.rdd.compress":  env.bool(x.upper() + '_SPARK_RDD_COMPRESS', default=False),
                "spark.serializer.objectStreamReset": env.int(x.upper() + '_SPARK_SERIALIZER_OBJECT_STREAM_RESET', default=ZERO),
                "spark.master": env.str(x.upper() + '_SPARK_MASTER', default=EMPTY_STRING),
                "spark.executor.id": env.str(x.upper() + '_SPARK_EXECUTOR_ID', default=EMPTY_STRING),
                "spark.driver.port": env.int(x.upper() + '_SPARK_DRIVER_PORT', default=ZERO),
                "spark.submit.deployMode": env.str(x.upper() + '_SPARK_SUBMIT_DEPLOY_MODE', default=EMPTY_STRING),
                "spark.app.name": env.str(x.upper() + '_SPARK_APP_NAME', default=EMPTY_STRING),
                "spark.ui.showConsoleProgress": env.bool(x.upper() + '_SPARK_UI_SHOW_CONSOLE_PROGRESS', default=True),

                # database
                "driver":  env.str(x.upper() + '_SPARK_DRIVER', default=EMPTY_STRING),
                "connector_type":  env.str(x.upper() + '_SPARK_CONNECTOR_TYPE', default=EMPTY_STRING),
                "db_type":  env.str(x.upper() + '_SPARK_DB_TYPE', default=EMPTY_STRING),
                "hostname":  env.str(x.upper() + '_SPARK_HOSTNAME', default=EMPTY_STRING),
                "port":  env.int(x.upper() + '_SPARK_PORT', default=ZERO),
                "username":  env.str(x.upper() + '_SPARK_DATABASE_USERNAME', default=EMPTY_STRING),
                "password": env.str(x.upper() + '_SPARK_DATABASE_PASSWORD', default=EMPTY_STRING),
                "database_name": env.str(x.upper() + '_SPARK_DATABASE_NAME', default=EMPTY_STRING),
                "database_options": env.str(x.upper() + '_SPARK_DATABASE_OPTIONS', default=EMPTY_STRING)
            }
        },
        "hive": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_SPARK', default=False),
            "type": "spark",
            "handler": HiveHandler,
            "settings": {
                "jarpath": env.str(x.upper() + '_HIVE_JAR_PATH', default=EMPTY_STRING),
                "url": env.str(x.upper() + '_HIVE_URL', default=EMPTY_STRING),
                "uid": env.str(x.upper() + '_HIVE_UID', default=EMPTY_STRING),
                "pwd": env.str(x.upper() + '_HIVE_PWD', default=EMPTY_STRING)
            }
        },
        "livy": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_SPARK', default=False),
            "type": "spark",
            "handler": LivyHandler,
            "settings": {
                "url": env.str(x.upper() + '_LIVY_URL', default=EMPTY_STRING),
                "username": env.str(x.upper() + '_LIVY_USERNAME', default=EMPTY_STRING),
                "password": env.str(x.upper() + '_LIVY_PASSWORD', default=EMPTY_STRING),
                "driver_memory": env.str(x.upper() + '_LIVY_DRIVER_MEMORY', default="512m"),
                "driver_cores": env.int(x.upper() + '_LIVY_DRIVER_CORES', default=1),
                "executor_cores": env.int(x.upper() + '_LIVY_EXECUTOR_CORES', default=1),
                "executor_memory": env.str(x.upper() + '_LIVY_EXECUTOR_MEMORY', default="512m"),
                "num_executors": env.int(x.upper() + '_LIVY_NUM_EXECUTORS', default=1),
                "queue": env.str(x.upper() + '_LIVY_QUEUE', default="default"),
                "name": env.str(x.upper() + '_LIVY_NAME', default="mlapp"),
                "heartbeat_timeout": env.int(x.upper() + '_LIVY_HEARTBEAT_TIMEOUT', default=60)
            }
        },
        # default rabbitmq settings for Kubernetes deployment. See ML App Wiki documentation
        "rabbitmq": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_MQ', default=False),
            "type": "message_queue",
            "handler": RabbitMQHandler,
            "settings": {
                'hostname': env.str(x.upper() + '_RABBITMQ_HOSTNAME', default=EMPTY_STRING),
                "port": env.int(x.upper() + '_RABBITMQ_PORT', default=ZERO),
                'username': env.str(x.upper() + '_RABBITMQ_USERNAME', default=EMPTY_STRING),
                'password': env.str(x.upper() + '_RABBITMQ_PASSWORD', default=EMPTY_STRING),
                'cert_path': env.str(x.upper() + '_RABBITMQ_CERT_PATH', default=EMPTY_STRING),
                'connection_timeout': env.int(x.upper() + '_RABBITMQ_CONNECTION_TIMEOUT', default=15)  # seconds
            }
        },
        "boto": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_FS', default=False),
            "handler": BotoStorageHandler,
            "type": "file_storage",
            "settings": {
                'endpoint': env.str(x.upper() + '_BOTO_ENDPOINT', default=EMPTY_STRING),
                'accessKeyId': env.str(x.upper() + '_BOTO_ACCESS_KEY', default=EMPTY_STRING),
                'secretKey': env.str(x.upper() + '_BOTO_SECRET_KEY', default=EMPTY_STRING),
                'is_secure': env.bool(x.upper() + '_BOTO_SECURE', default=False)
            }
        },
        "ibm_boto3": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_FS', default=False),
            "handler": IBMBoto3StorageHandler,
            "type": "file_storage",
            "settings": {
                'endpoint': env.str(x.upper() + '_IBM_BOTO3_ENDPOINT', default=EMPTY_STRING),
                'api_key_id': env.str(x.upper() + '_IBM_BOTO3_API_KEY_ID', default=EMPTY_STRING),
                'service_crn': env.str(x.upper() + '_IBM_BOTO3_SERVICE_CRN', default=EMPTY_STRING)
            }
        },
        "kafka": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_MQ', default=False),
            "handler": KafkaHandler,
            "type": "message_queue",
            "settings": {
                'hostname': env.str(x.upper() + '_KAFKA_HOSTNAME', default=EMPTY_STRING),
                "port": env.int(x.upper() + '_KAFKA_PORT', default=ZERO),
                # 'listen_queues_names': env.list(x.upper() + '_KAFKA_LISTEN_QUEUES_NAMES', default=[]),
                # 'send_queue_names': env.list(x.upper() + '_KAFKA_SEND_QUEUE_NAMES', default=[]),
                'connection_timeout': env.int(x.upper() + '_KAFKA_CONNECTION_TIMEOUT', default=15),  # seconds
                "username": env.str(x.upper() + '_KAFKA_USERNAME', default=EMPTY_STRING),
                "password": env.str(x.upper() + '_KAFKA_PASSWORD', default=EMPTY_STRING),
            }
        },
        "kafka_kerberos": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_MQ', default=False),
            "handler": KafkaKerberosHandler,
            "type": "message_queue",
            "settings": {
                "hostnames": env.str(x.upper() + '_KAFKA_HOSTNAMES', default=EMPTY_STRING),
                "group_id": env.str(x.upper() + '_KAFKA_GROUP_ID', default=EMPTY_STRING),
                "keytab": env.str(x.upper() + '_KAFKA_KEYTAB', default=EMPTY_STRING),
                "ca_location": env.str(x.upper() + '_KAFKA_CA_LOCATION', default=EMPTY_STRING),
                "username": env.str(x.upper() + '_KAFKA_USERNAME', default=EMPTY_STRING),
                "domain_name": env.str(x.upper() + '_KAFKA_DOMAIN_NAME', default=EMPTY_STRING)
            }
        },
        "azureml_queue": lambda x: {
            "main": env.bool(x.upper() + '_MAIN_MQ', default=False),
            "handler": AMLQueue,
            "type": "message_queue",
            "settings": {
                'experiment_name': env.str(x.upper() + '_AML_EXPERIMENT_NAME', default=EMPTY_STRING),
            }
        }
    }


settings = {
    # assets output is saved locally at the path below.
    "local_storage_path": "output",
    "latest_file_name": "latest_ids.json",

    # assets output is fallback at the path below.
    "temporary_storage_path": "temporary_output",

    # if file storage configurations are set, pipeline will save output there instead of locally
    "file_store_buckets": {
        'objects': 'mlapp-objects',
        'csvs': 'mlapp-csvs',
        'configs': 'mlapp-configs',
        'metadata': 'mlapp-metadata',
        'imgs': 'mlapp-imgs',
        'logs': 'mlapp-logs'
    },

    # main queues for app to send/listen
    "queues": {
        'listen_queue_names': ['analysis_general_listen'],
        'send_queue_name': 'analysis_respond',
        'analysis_logs_queue_name': 'analysis_logs',
    },
    "pipelines": {
        'train': ['load_train_data', 'clean_train_data', 'transform_train_data', 'train_model'],
        'explore_data': ['load_train_data', 'clean_train_data', 'transform_train_data', 'visualization'],
        'feature_engineering': ['load_train_data', 'clean_train_data', 'transform_train_data', 'cache_features'],
        'reuse_features_and_train':['load_features','train_model'],
        'forecast': ['load_forecast_data', 'clean_forecast_data', 'transform_forecast_data', 'forecast'],
        'predictions_accuracy': ['load_actuals_data', 'update_actuals', 'evaluate_prediction_accuracy'],
        'retrain': ['load_train_data', 'clean_train_data', 'transform_train_data', 'train_model'],
        'train_flow': ['load_train_data', 'clean_train_data', 'transform_train_data', 'cache_features', 'train_model'],
        'forecast_flow': ['load_forecast_data', 'clean_forecast_data', 'transform_forecast_data', 'forecast']
    },
    # dynamically configured
    "services": {}
}



import multiprocessing
from mlapp.env_loader import EnvironmentLoader
from mlapp.utils.general import read_json_file
import importlib
import uuid
import json
from mlapp.handlers.wrappers.database_wrapper import database_instance
from mlapp.handlers.wrappers.file_storage_wrapper import file_storage_instance
from mlapp.handlers.wrappers.message_queue_wrapper import message_queue_instance
from mlapp.handlers.wrappers.spark_wrapper import spark_instance
from mlapp.managers.flow_manager import FlowManager
import traceback
from ast import literal_eval
from mlapp.config import settings, environment_services
import os


class MLApp(object):
    MLAPP_SERVICE_TYPE = '_MLAPP_SERVICE_TYPE'

    def __init__(self, inner_settings=None):
        """
        Constructor for the MLApp Class.
        This class, when becomes instantiated is the main endpoint for the ML App Library.
        To this constructor you pass in your custom settings, which sets up your customized configuration of ML App,
        and Environment that you defined with the ML App CLI.
        After the instantiation you can use the instance to run:
         - Flows (single-process and multi-process)
         - Applications/Workers (using the queue listener)
         - Send configurations to an outside queue
        """
        if inner_settings is None:
            inner_settings = {}

        for key in inner_settings:
            if isinstance(inner_settings[key], dict):
                if key not in settings:
                    settings[key] = {}
                settings[key].update(inner_settings[key])
            elif isinstance(inner_settings[key], list):
                if key not in settings:
                    settings[key] = []
                settings[key] += inner_settings[key]
            else:
                settings[key] = inner_settings[key]

        # init environment services
        env = EnvironmentLoader.load(filename=inner_settings.get('env_file_path', ''))
        env_services_dict = {
            k.replace(self.MLAPP_SERVICE_TYPE, ''): os.environ[k].lower()
            for k in os.environ if k.endswith(self.MLAPP_SERVICE_TYPE)
        }
        settings['services'] = EnvironmentLoader.create_services(env, env_services_dict, environment_services)
        for wrapper_instance in [file_storage_instance, database_instance, message_queue_instance, spark_instance]:
            wrapper_instance.init()

    # ======== TASK RUN  ==========
    def _on_callback(self, message_body):
        """
        This is the function that executes the configuration sent to the application/worker.
        :param message_body: the configuration in bytes/string format.
        :return: None
        """
        job_id = 'None'
        try:
            message_body = message_body.decode("utf-8")
        except AttributeError as attrError:
            pass    # message body is string and not bytes

        print("Hello, The following task is consumed:   " + str(message_body))
        try:
            message_config = json.loads(literal_eval(message_body))
        except Exception as first_error:
            try:
                message_config = json.loads(message_body)
            except Exception as second_error:
                print("Error response: " + str('Message not in JSON format'))
                traceback.print_exc()
                self._send_error_response_to_mq(job_id, '-1', str('Message not in JSON format'))
                return

        print(message_config)
        try:
            job_id = str(message_config.get('job_id', str(uuid.uuid4())))
            results = self._run_flow(job_id, message_config)
            self._send_ok_response_to_mq(
                job_id, results.get('status_code', -1), 'all went ok', results.get('response', {}))

        except Exception as error:
            print("Error response: " + str(error))
            traceback.print_exc()
            self._send_error_response_to_mq(job_id, '-1', str(error))
        finally:
            pass

    # =========== TASK HANDLERS ===========
    def _run_flow(self, job_id, config):
        """
        This is the function that executes the Flow of your configuration.
        :param job_id: the job identifier used for monitoring via the Control Panel.
        :param config: the configuration as Dictionary.
        :return: Dictionary containing the status and response of the flow run.
        """
        # update job
        database_instance.update_job_running(job_id)
        # call Flow_manager to run the job
        status_code, response, _ = FlowManager(job_id, config).run()

        return {'status_code': status_code, 'response': response}

    # ======== MQ HANDLERS =========
    def _send_ok_response_to_mq(self, job_id, status_code, status_msg, result):
        """
        This function sends response back to the Control Panel via queue if the job succeeded
        :param job_id: the job identifier used for monitoring via the Control Panel.
        :param status_code: result status of the flow run.
        :param status_msg: result message of the flow run.
        :param result: response of the flow run - if json serialized returned in the message queue as well.
        :return: None
        """
        response_obj = {
            "job_id": job_id, "status_code": status_code, "status_msg": status_msg
        }
        try:
            # trying to JSON-ify result object
            response_obj['result'] = result
            response_json = json.dumps(response_obj)
        except Exception as error:
            print(error)
            response_obj['result'] = {}
            response_json = json.dumps(response_obj)

        message_queue_instance.send_message(settings['queues']['send_queue_name'], response_json)

    def _send_error_response_to_mq(self, job_id, status_code, status_msg):
        """
        This function sends response back to the Control Panel via queue if the job failed
        :param job_id: the job identifier used for monitoring via the Control Panel.
        :param status_code: error status of the flow run.
        :param status_msg: error message of the flow run.
        :return: None
        """
        response_json = json.dumps({"job_id": job_id, "status_code": status_code, "status_msg": status_msg})
        message_queue_instance.send_message(settings['queues']['send_queue_name'], response_json)

    def _dispatch_jobs_to_mq(self, configurations):
        """
        This function sends configurations to the queue to be picked up later by a listening Application/Worker.
        :param configurations: list of configurations to be sent
        :return: None
        """
        for configuration in configurations:
            response_json = json.dumps(configuration)
            message_queue_instance.send_message(settings['queues']['send_queue_name'], json.dumps(response_json))

    # ======== LISTENER  =========
    def run_listener(self):
        """
        This function is an endpoint of the ML App Library to be used in an Application/Worker.
        It sets up a listening queue indefinitely waiting for configuration to process upon receive.
        """
        message_queue_instance.listen_to_queues(settings['queues']['listen_queue_names'], self._on_callback)

    # ======== RUN CONFIG  =========
    def run_flow(self, asset_name, config_path, config_name=None, **kwargs):
        """
        This function is an endpoint of the ML App Library to be used in a local environment.
        It runs a local configuration file in your local computer.
        :param asset_name: name of the asset to be run
        :param config_path: path to configuration file
        :param config_name: in case configuration file is python looks for variable in this name as the configuration
        """
        job_id = str(uuid.uuid4())
        try:
            config = read_json_file(config_path)
        except Exception as err:
            config = self._read_py_file(asset_name, config_path, config_name)
        self._insert_latest_id_in_config(config)
        _, run_ids, outputs = FlowManager(job_id, config, **kwargs).run()
        self._update_latest_model_id(config, run_ids)

    @staticmethod
    def run_flow_from_config(config):
        return FlowManager("deployment", config).run()

    # ======== SEND CONFIG TO MQ  =========
    def run_msg_sender(self, asset_name, config_path, config_name=None):
        """
        This function is an endpoint of the ML App Library to be used in a local environment.
        It sends a local configuration file in your local computer to be run in an outside Application/Worker via
        message queue.
        :param asset_name: name of the asset to be run
        :param config_path: path to configuration file
        :param config_name: in case configuration file is python looks for variable in this name as the configuration
        """
        try:
            message_to_send = read_json_file(config_path)
        except Exception as e:
            message_to_send = self._read_py_file(asset_name, config_path, config_name)

        job_id = str(uuid.uuid4())
        message_to_send['job_id'] = job_id
        message_queue_instance.send_message(settings['queues']['listen_queue_names'][0], json.dumps(message_to_send))
        print("Message Sent (job_id: " + job_id + "): ", asset_name, config_path)

    # ======== RUN CONFIGS MULTIPROCESSING  =========
    def run_configs_multiprocessing(self, instructions):
        """
        This function is an endpoint of the ML App Library.
        It runs multiple configurations in multi-processing.
        :param instructions: list of instruction to send to each process
        """
        jobs = []
        for instruction in instructions:
            p = multiprocessing.Process(target=self._run_config_multiprocess, args=(instruction,))
            jobs.append(p)
            p.start()

        for p in jobs:
            p.join()

    # ======== HELPER PRIVATE FUNCTIONS  =========
    def _run_config_multiprocess(self, instruction):
        """
        This function is executes instruction of a process when used by `run_configs_multiprocessing`.
        :param instruction: instruction Dictionary containing asset_name, config_path and config_name.
            - asset_name: name of the asset to be run
            - config_path path to configuration file
            - config_name in case configuration file is python looks for variable in this name as the configuration
        """
        try:
            self.run_flow(instruction['asset_name'], instruction['config_path'], instruction.get('config_name'))
        except Exception as err:
            print(err)
            traceback.print_exc()

    @staticmethod
    def _read_py_file(asset_name, config_path, config_name):
        """
        This function fetches a configuration Dictionary stored in a python file.
        :param asset_name: name of the asset to be run
        :param config_path: path to configuration file
        :param config_name: variable in the python file containing the configuration
        :return: Configuration as a Dictionary
        """
        spec = importlib.util.spec_from_file_location(asset_name, config_path)
        config = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(config)
        return config.__dict__[config_name]

    @staticmethod
    def _insert_latest_id_in_config(config):
        """
        This is a helper function for using `latest` feature in local environment.
        Updates current configuration to be run with the latest id stored in a local file containing it
        by reference of asset name.
        :param config: current flow run configuration as a Dictionary
        """
        # prepare latest file
        local_path = settings.get('local_storage_path', 'output')
        latest_file_name = settings.get('latest_file_name', 'latest_ids.json')
        latest_ids_path = os.path.join(local_path, latest_file_name)
        try:
            with open(latest_ids_path) as f:
                latest = json.load(f)
        except:
            latest = {}

        # iterate pipelines
        for i, pipeline in enumerate(config.get('pipelines_configs', [])):
            # iterate optional ids
            for id_type in ['model_id', 'data_id', 'reuse_features_id']:
                # check if requested latest
                if pipeline.get('job_settings', {}).get(id_type, None) == 'latest':
                    # get current asset name
                    asset_name = pipeline['job_settings']['asset_name']
                    # check if available id
                    if asset_name in latest:
                        # TODO: add here asset label level
                        # TODO: add here data_id/model_id/reuse_features_id
                        config['pipelines_configs'][i]['job_settings'][id_type] = latest[asset_name]
                    else:
                        # raise exception as not found id
                        raise Exception("Could not find latest `" + id_type + "` for `" + asset_name + "`. \n"
                                        "Please update your config with a valid `" + id_type + "`")

    @staticmethod
    def _update_latest_model_id(config, run_ids):
        """
        This is a helper function for using `latest` feature in local environment.
        Updates local file containing the latest id used for an asset.
        :param config: current flow run configuration as a Dictionary
        :params run_ids: list of mlapp identifiers generated in the current flow run.
        """
        # prepare latest file
        local_path = settings.get('local_storage_path', 'output')
        latest_file_name = settings.get('latest_file_name', 'latest_ids.json')
        if not os.path.exists(local_path):
            os.makedirs(local_path)
        latest_ids_path = os.path.join(local_path, latest_file_name)

        latest = {}
        try:
            with open(latest_ids_path) as f:
                latest = json.load(f)
        except:
            pass

        # iterate over pipelines
        for pipeline, run_id in zip(config['pipelines_configs'], run_ids):
            # check if ran any pipeline where id is being stored
            # TODO: add here asset label level
            # TODO: add here data_id/model_id/reuse_features_id
            if pipeline['job_settings']['pipeline'] in ['train', 'feature_engineering']:
                latest[pipeline['job_settings']['asset_name']] = run_id

            with open(latest_ids_path, 'w') as f:
                json.dump(latest, f)



from numpy import random
from assets.classification.classification_visualizations import plot_multiple_confusion_matrix
from mlapp.managers import ModelManager, pipeline
from sklearn.model_selection import StratifiedKFold, train_test_split
from mlapp.utils.automl import AutoMLPandas


class ClassificationModelManager(ModelManager):
    @pipeline
    def train_model(self, data):
        """
        This function should create a model  and train the model with the features generated by the function build_and_store_features in the Task_Manager Class
        :param data: pandas DataFrame that contains X, y for training the model
        :return local_model_path: string: pickle file path
        :return img_file_names: list of strings: of img file names paths
        :return model_results: dict: contains info of model such as accuracy, configs, etc.
        :return predictions_df: pandas DataFrame: estimates+trend+seasonality of the train period
        """
        # -------------------------------- preparation for train --------------------------------
        variable_to_predict = self.model_settings['variable_to_predict']

        # Sampling the data
        data = self._sample_data(data)

        # Preparing the data
        X_train, X_test, y_train, y_test = train_test_split(
            data.drop(variable_to_predict, axis=1), data[variable_to_predict],
            train_size=self.model_settings.get('train_percent', 0.8), stratify=data[variable_to_predict])

        # Run AutoML
        auto_ml_config = self.model_settings.get('auto_ml', {})
        estimator_family = auto_ml_config.get('estimator', 'binary')
        cv_splitter = StratifiedKFold(n_splits=5, shuffle=True, random_state=random.seed(0))
        result = AutoMLPandas(
            estimator_family, **auto_ml_config.get(estimator_family, {}),
            feature_selection=auto_ml_config.get('feature_selection')).run(
            X_train, y_train, X_test, y_test, cv=cv_splitter)

        # Print report
        result.print_report(full=False)

        # Save results
        self.save_automl_result(result)

        # Add confusion matrix plot
        self.save_image('confusion_matrix', plot_multiple_confusion_matrix(
            result.get_train_predictions(), y_train, result.get_test_predictions(), y_test))

        # Save probability thresholds
        proba_threshold = result.predict_proba_by_threshold(
            X_train, X_test, y_train, y_test, self.model_settings.get('predict_proba_threshold', 0.5))
        self.save_metadata('predict_proba_by_threshold', proba_threshold)

        # Add predictions
        self.add_predictions(X_train.index, result.get_train_predictions(), y_train, 'TRAIN')
        self.add_predictions(X_test.index, result.get_test_predictions(), y_test, 'TEST')

    @pipeline
    def forecast(self, data_df):
        result = self.get_automl_result()

        # get best model
        model = result.get_best_model()

        # get selected features from feature selection algorithm
        selected_features_names = result.get_selected_features()

        # filter data according to selected features
        filtered_data = data_df[selected_features_names]

        # predicting
        predictions = model.predict(filtered_data)

        self.add_predictions(primary_keys_columns=data_df.index, y_hat=predictions, y=[], prediction_type='forecast')

    def _sample_data(self, data):
        sample_data = self.model_settings.get('down_sample_method', None)
        if sample_data is not None and sample_data.get('flag', False):
            n_samples = sample_data.get('n_samples', data.shape[0])
            seed = sample_data.get('seed', 1500)
            if isinstance(n_samples, int) and n_samples <= data.shape[0]:
                return data.sample(n=n_samples, random_state=seed)
            elif isinstance(n_samples, float) and (0.0 < n_samples <= 1.0):
                n_samples = int(data.shape[0] * n_samples)
                return data.sample(n=n_samples, random_state=seed)
            else:
                print('Warning: "n_samples" can be an Integer Type that <= number of data samples or Float type (ratio)'
                      ' that bigger than 0 and smaller or equal to 1 (0 < n_sample <= 1) \n Taking Full Data!')
            print('After data sampling: (' + repr(data.shape[0]) + ',' + repr(data.shape[1]) + ')')

        return data

    def refit(self, data):
        raise NotImplementedError()



import os
import logging
import shutil
import imghdr
import glob
from mlapp.handlers.file_storages.file_storage_interface import FileStorageInterface
from mlapp.integrations.aml.utils.run_class import get_model_register_name
from mlapp.integrations.aml.utils.constants import OUTPUTS_FOLDER, AML_MLAPP_FOLDER
from mlapp.utils.general import read_json_file
from azureml.core import Model, Run
from azureml.exceptions import RunEnvironmentException
from mlapp.utils.exceptions.framework_exceptions import SkipServiceException


class AmlModelStorageHandler(FileStorageInterface):
    def __init__(self, settings):
        """
        Initializes the AmlModelStorageHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(AmlModelStorageHandler, self).__init__()

        self.registered_models = {}
        self.downloaded_files = {}
        self.temporary_storage = settings.get('temporary_storage_path')

        self.imgs_bucket = settings.get('file_store_buckets', {}).get('imgs', 'mlapp-imgs')
        self.metadata_bucket = settings.get('file_store_buckets', {}).get('metadata', 'mlapp-metadata')

        try:
            self.run = Run.get_context(allow_offline=False)
        except RunEnvironmentException as e:
            raise SkipServiceException('Skip AmlModelStorageHandler handler')
        self.ws = self.run.experiment.workspace

        # create AML outputs folder
        os.makedirs(OUTPUTS_FOLDER, exist_ok=True)
        os.makedirs(os.path.join(OUTPUTS_FOLDER, AML_MLAPP_FOLDER), exist_ok=True)

    def _register(self, run_id, asset_name, asset_label=None):
        """
        This method register a new model as a AML reference.
        :param run_id: run's identifier
        :param asset_name: name of asset of the current run
        :param asset_label: name of label of the current run
        :return: None
        """
        tags = {
            "run_id": run_id,
            "asset_name": asset_name
        }

        if asset_label is not None:
            tags["asset_label"] = asset_label

        register_path = os.path.join(OUTPUTS_FOLDER, AML_MLAPP_FOLDER)
        Model.register(
            self.ws,
            model_path=register_path,
            model_name=get_model_register_name(run_id),
            tags=tags,
            description=asset_name)

    def download_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        This method downloads a file into your machine from AML datastore
        :param bucket_name:  name of the bucket/container, unnecessary parameter
        :param object_name: name of the object/file
        :param file_path: path to local file, unnecessary parameter
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        :return: None
        """
        # get run_id from object_name
        current_id = object_name.split('_')[0]

        if current_id not in self.downloaded_files.keys() or not self.downloaded_files[current_id]:
            # load registered models
            model_name = get_model_register_name(current_id)
            model = Model(self.ws, name=model_name)
            model.download(target_dir=current_id, exist_ok=True)

            # mark current_id as registered
            self.downloaded_files[current_id] = {
                "path": current_id
            }

        # move file to AML_MLAPP_FOLDER folder from current_id folder
        if os.path.exists(os.path.join(current_id, AML_MLAPP_FOLDER, object_name)):
            shutil.move(os.path.join(current_id, AML_MLAPP_FOLDER, object_name), self.temporary_storage)

    def upload_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        This method move a file to the registered folder of AML
        :param bucket_name:  name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        :return: None
        """
        # copy file from file_path to register_path. file deletion is handled by job manager
        register_path = os.path.join(OUTPUTS_FOLDER, AML_MLAPP_FOLDER)
        output_path = os.path.join(register_path, os.path.basename(file_path))
        shutil.copy(file_path, output_path)

        run_id = kwargs.get('run_id')
        if run_id not in self.registered_models.keys():
            # save run_id parameters.
            self.registered_models[run_id] = {
                "asset_name": kwargs.get('asset_name'),
                "asset_label": kwargs.get('asset_label')
            }

        # log images
        if bucket_name == self.imgs_bucket:
            # check if file is an image
            is_img = imghdr.what(file_path)
            if is_img is not None:
                file_name = os.path.splitext(os.path.basename(file_path))[0]
                image_name = '_'.join(file_name.split('_')[1:])
                self.run.log_image(name=image_name, path=output_path)
        # log metadata
        else:
            if bucket_name == self.metadata_bucket:
                model_metadata = read_json_file(file_path)
                scores = model_metadata.get('models', {}).get('scores', None)
                if scores is not None:
                    for score_key, score_value in scores.items():
                        self.run.log(name=score_key, value=score_value)

    def list_files(self, bucket_name, prefix="", *args, **kwargs):
        """
        Lists files in file storage
        :param bucket_name: name of the bucket/container
        :param prefix: prefix string to search by
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        try:
            # get run_id from object_name
            current_id = prefix

            if current_id not in self.downloaded_files.keys():
                # load registered models
                model_name = get_model_register_name(current_id)
                model = Model(self.ws, name=model_name)
                model.download(target_dir=current_id, exist_ok=True)

                # mark current_id as registered
                self.downloaded_files[current_id] = {
                    "path": current_id
                }

            # get files from current_id folder and temporary_storage folder
            list_files = os.listdir(os.path.join(current_id, AML_MLAPP_FOLDER))
            list_files_extra = list(
                map(lambda f: os.path.basename(f), glob.glob(os.path.join(self.temporary_storage, current_id + "_*"))))
            list_files.extend(list_files_extra)
            return list_files

        except Exception as e:
            logging.error(e)
            raise e

    def _clear(self):
        """
        This method deletes all used and unnecessary files.
        :return: None
        """
        # deletes all unnecessary files
        for run_id_to_del, metadata in self.downloaded_files.items():
            path = metadata.get("path", None)
            if path is not None and os.path.isdir(path):
                shutil.rmtree(path)

        # clear self.downloaded_files
        self.downloaded_files = {}

    def postprocess(self):
        """
        This method register and clear unnecessary files.
        :return: None
        """

        # register models
        for run_id, values in self.registered_models.items():
            asset_name = values.get('asset_name')
            asset_label = values.get('asset_label')
            self._register(run_id, asset_name, asset_label)

        # clear all files after registering
        self._clear()



import pickle
import zipfile
import tempfile
import glob
from pathlib import Path
import numpy as np
import shutil
import os
import json
import collections
from ast import literal_eval
import collections.abc
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas


def recursive_dict_update(d, u):
    if hasattr(u, 'items'):
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                d[k] = recursive_dict_update(d.get(k, {}), v)
            else:
                d[k] = v
        return d


def search_key_recursive( d, key):
    if hasattr(d, 'items'):
        for k, v in d.items():
            if key == k:
                yield v
            if isinstance(v, dict):
                for result in search_key_recursive(v, key):
                    yield result
            elif isinstance(v, list):
                for i in v:
                    for result in search_key_recursive(i, key):
                        yield result


def __convert__(data):
    if isinstance(data, str):
        return str(data)
    elif isinstance(data, collections.abc.Mapping):
        return dict(map(__convert__, data.items()))
    elif isinstance(data, collections.abc.Iterable):
        return type(data)(map(__convert__, data))
    else:
        return data


def read_json_file(path, file_name=None):
    if file_name is None:
        with open(path, "r") as f:
            data = json.load(f)
    else:
        with open(os.path.join(path, file_name), "r") as f:
            data = json.load(f)
    return __convert__(data)


def load_json_unicode(body):
    data = json.loads(literal_eval(body))
    return data


class NumpyEncoder(json.JSONEncoder):
    """
    Special json encoder for numpy types
    """
    def default(self, obj):
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8, np.int16, np.int32,
                            np.int64, np.uint8, np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def get_project_root():
    # change this path accordingly when moving this file
    return Path(__file__).parent.parent.parent


def load_pickle_to_object(path):
    obj = pickle.load(open(path, 'rb'))
    return obj


def save_object_to_pickle(model, path):
    pickle.dump(model, open(path, 'wb'))


def extract_zip_file(file_path):
    try:
        zip_ref = zipfile.ZipFile(file_path, 'r')
        zip_ref.extractall(file_path)
        zip_ref.close()
    except Exception as e:
        raise RuntimeError('failed to unzip file: ' + str(e))


def compress_to_zip_file(file_path):
    try:
        zip_ref = zipfile.ZipFile(file_path + ".zip", 'w', zipfile.ZIP_DEFLATED)
        zip_ref.write(file_path)
        zip_ref.close()
    except Exception as e:
        raise RuntimeError('failed to compress file: ' + str(e))


def save_object_tempfile(file_name, obj):
    tempdir = tempfile.mkdtemp()
    file_path = os.path.join(tempdir, file_name + '.pkl')
    with open(file_path, 'wb') as fp:
        pickle.dump(obj, fp)
    return file_path


def get_compressed_file_and_remove_uncompressed(file_path):
    compress_to_zip_file(file_path)
    os.remove(file_path)


def compress_folder_to_zip(file_path):
    shutil.make_archive(file_path, 'zip', file_path)
    shutil.rmtree(file_path)


def uncompress_zip_to_folder(file_path, extract_dir):
    if not os.path.exists(extract_dir):
        os.mkdir(extract_dir)
        shutil.unpack_archive(file_path, extract_dir=extract_dir)
    os.remove(file_path)


def save_object_as_json(obj, file_path):
    with open(file_path, 'w') as outfile:
        json.dump(obj, outfile, cls=NumpyEncoder)


def is_jsonable(obj):
    """
    checks if object is json serializable.
    :param obj: python object
    :return: bool
    """
    try:
        json.dumps(obj,cls=NumpyEncoder)
        return True
    except:
        return False


def update_json_file(new_data, file_path):
    """
    This function reads existing data from json file and append it with the new data and save the file.
    :param new_data: dictionary - new data to append
    :param file_path: string - path to file
    :return: None
    """

    # read existing data from json file
    with open(file_path, 'r') as f:
        data = json.load(f)

    # updates the json file data
    data.update(new_data)

    # write new data to the json file
    with open(file_path, 'w') as f:
        json.dump(data, f)


def validate_json(d, warnning_message=None):
    """
    This function get an object and check if its a vaild json
    and deletes everything that not json serializable
    :param d: python dictionary
    :param warnning_message: str
    :return: None (void)
    """
    if not isinstance(d, dict):
        raise Exception('Error: object is not dictionary')

    keys_to_delete = []
    for key,value in d.items():

        # if d[key] is also a dictionary call recursive check
        if isinstance(d[key], dict):
            validate_json(d[key], warnning_message)

        if not is_jsonable(value):
            keys_to_delete.append(key)

    for key in keys_to_delete:
        if key in d:
            if warnning_message is not None:
                print("Warnning: " + warnning_message % (str(key)))
            del d[key]


def save_figure_to_png(figure, path, file_name):
    try:
        figure.savefig(path)
    except:
        try:
            FigureCanvas(figure)
            figure.savefig(path)
        except Exception as e:
            raise Exception('Failed to save %s plot image.' % (str(file_name)))


def read_files_names_from_local_storage(prefix, local_storage_path):
    '''
    Reads all files starting with prefix from local storage.
    @param prefix: str, filename name prefix to look for
    @param local_storage_path: str, local path to look for files.
    @return: list of strings, file names.
    '''
    return list(map(lambda f: os.path.basename(f), glob.glob(os.path.join(local_storage_path, prefix + "_*"))))


def create_tempdir(name):
    return tempfile.mkdtemp(prefix=name)


def delete_directory_with_all_contents(path):
    shutil.rmtree(path=path)


def create_directory(directory_name, path=''):
    # concat the path and directory name.
    full_directory_path = os.path.join(path, directory_name)
    full_directory_path = os.path.join(os.getcwd(), full_directory_path)

    # checks if directory already exists, if not then creating it.
    if not os.path.exists(full_directory_path):
        os.makedirs(full_directory_path)
        return full_directory_path
    else:
        raise Exception('ERROR: path %s already exists.' % str(full_directory_path))




import os, json, socket, shutil, re


def check_port_in_use(port, ip='127.0.0.1'):
    '''
    This function check if port is in use under some if
    :param port: int
    :param ip: string
    :return: bool
    '''
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip, int(port)))
        s.shutdown(1)
        return True
    except:
        return False


def get_available_port(port='8081', ip='127.0.0.1'):
    '''
    This function return free port
    :param port: int
    :param ip: string
    :return: int
    '''
    while True:
        is_free = check_port_in_use(port, ip)
        if is_free:
            break
        else:
            port += 1
    return port


def str_to_camelcase(s):
    '''
    This function takes model_names, that contains underscore and transforms them to camelcase.
    :param s: string
    :return: string (camelcase)
    '''
    res = ''.join(x for x in s.title() if not x.isspace())
    res = res.replace("_", "")
    return res


def validate_str(s):
    '''
    This function validate a string by the rules:
    * turn each underscore to dash.
    * transform all string characters to uppercase.
    * remove spaces
    :param s: string
    :return: string
    '''
    new_s = s
    new_s = new_s.replace("_", "-")
    new_s = ''.join(x for x in new_s.title() if not x.isspace())
    new_s = new_s.upper()
    return new_s


def create_file(file_name, path='', permissions='w+', content=''''''):
    try:

        # concat the path and filename.
        full_file_path = os.path.join(path, file_name)
        full_file_path = os.path.join(os.getcwd(), full_file_path)

        # create file according to parameters.
        f = open(full_file_path, permissions)

        # write content in file.
        if content != '''''':
            f.write(content)

        # close file.
        f.close()

    except Exception as e:
        raise e


def write_to_file(file_name, path='', content=''''''):
    try:

        # concat the path and filename.
        full_file_path = os.path.join(path, file_name)
        full_file_path = os.path.join(os.getcwd(), full_file_path)

        # write content into file.
        with open(full_file_path, "a") as f:
            f.write(content)

    except Exception as e:
        raise e


def create_directory(directory_name, path='', include_init=True):
    try:

        # concat the path and directory name.
        full_directory_path = os.path.join(path, directory_name)
        full_directory_path = os.path.join(os.getcwd(), full_directory_path)

        # checks if directory already exists, if not then creating it.
        if not os.path.exists(full_directory_path):
            os.makedirs(full_directory_path)

        if include_init is True:
            create_file(file_name='__init__.py', path=full_directory_path)

    except Exception as e:
        raise e


def set_env(env_filename, env_path='env'):
    try:
        with open(os.path.join(os.getcwd(), 'config.py'), 'r') as f:
            config_content = f.read()
            exec(config_content)
            d = eval("settings")

        if env_filename == '':
            d['env_file_path'] = env_filename
        else:
            d['env_file_path'] = os.path.join(env_path, env_filename)

        config_file_content = '''settings = ''' + json.dumps(d, indent=2)
        create_file(file_name='config.py', content=config_file_content)

    except Exception as e:
        raise e


def get_env():
    try:
        with open(os.path.join(os.getcwd(), 'config.py'), 'r') as f:
            config_content = f.read()
            exec(config_content)
            d = eval("settings")

        env_filename = d.get('env_file_path', None)

        if env_filename is not None:
            return env_filename
        else:
            raise Exception("Env file not found, please create it before!")

    except Exception as e:
        raise e


def format_service_dictto_txt(service_dict):
    '''

    :param service_dict:
    :return:
    '''
    text = ''''''
    for key, value in service_dict.items():
        str_val = str(value)
        if str_val != "":
            text += key + '''=''' + str_val + '''\n'''

    return text


def check_for_service_uniqueness_name(service_name, env_filename):
    '''
    checks is service_name not exists in services
    :param service_name: string
    :return: bool
    '''

    try:
        # find env file
        env_full_path = os.path.join(os.getcwd(), env_filename)

        services = {}
        with open(env_full_path) as f:
            lines = f.readlines()
            for line in lines:
                line_striped = line.strip()

                if len(line_striped) == 0:
                    continue

                line_split = line_striped.split('_')
                if len(line_split) > 0:
                    current_service_name = line_split[0]
                    if current_service_name != '' and current_service_name[0] != '#':
                        services[current_service_name.lower()] = True

        return not service_name.lower() in services.keys()

    except Exception as e:
        raise e


def to_lower(s):
    return s.lower()


def to_upper(s):
    return s.upper()


def clean_spaces(s):
    return s.replace(" ", "")


def is_int(s):
    try:
        int(s)
        return True
    except:
        return False


def is_positive(s):
    try:
        return int(s) > 0
    except:
        return False


def to_int(s):
    return int(s)


def list_one_check(s):
    l = s.split(',')
    l = list(filter(lambda x: x != '', l))
    res = ''
    for x in l:
        res += x + ','
    return res[:-1]


def copy_files(src, dest, except_list=None, include_list=None):
    if except_list is None:
        except_list = []
    try:
        if not os.path.exists(src) or not os.path.exists(dest):
            raise Exception("src folder and destination folder must be exists")

        src_files = os.listdir(src)

        if include_list is not None:
            src_files = [file for file in src_files if file in include_list]

        for file_name in src_files:
            if file_name in except_list:
                continue
            else:
                full_filename = os.path.join(src, file_name)
                if os.path.isfile(full_filename):
                    shutil.copy(full_filename, dest)
                elif os.path.isdir(full_filename):
                    copy_files(os.path.join(src, file_name), os.path.join(dest, file_name), except_list)

    except Exception as e:
        raise e


def check_model_name(s):
    underscore = '_'
    dash = '-'

    if underscore in s and dash in s:
        return [underscore, dash]
    elif underscore in s:
        return [underscore]
    elif dash in s:
        return [dash]
    else:
        raise Exception('Name can include only letters, number, underscore and dash character.')


def str_capitalize(s):
    try:
        words = re.split(r"[, \-!#$%^&*@?_:]+", s)
        words = map(lambda x: x.capitalize(), words)
        res = "".join(words)
        return res
    except Exception as e:
        raise e



from mlapp.handlers.instance import spark_handler
from mlapp.managers import DataManager, pipeline
import pyspark.sql.functions as F
from pyspark.ml.feature import Bucketizer


class SparkRegressionDataManager(DataManager):
    @pipeline
    def load_train_data(self, *args):
        return self._load_data()

    @pipeline
    def load_forecast_data(self,*args):
        return self._load_data()

    @pipeline
    def clean_train_data(self, data):
        return self._clean_data(data)

    @pipeline
    def clean_forecast_data(self, data):
        missing_values = self.get_metadata('missing_values', {})
        return self._clean_data(data, missing_values)

    @pipeline
    def transform_train_data(self, data):
        return self._transform_data(data)

    @pipeline
    def transform_forecast_data(self, data):
        return self._transform_data(data)

    @pipeline
    def load_target_data(self, *args):
        raise NotImplementedError()

    # ------------------------- private functions for load/clean/transform ------------------------------

    def _load_data(self):
        return spark_handler('LOCAL-SPARK').load_csv_file(self.data_settings["local_file_path"], inferSchema=True)

    def _clean_data(self, data, stored_missing_values=None):
        missing_values = {}

        data_handling = self.data_settings.get('data_handling', {})
        features_handling = data_handling.get('features_handling', {})

        # remove features by null percentage
        null_percentage = data_handling.get("feature_remove_by_null_percentage", 0.5)
        null_percentages = data.select(
            [(F.count(F.when(F.isnull(c), c)) / data.count()).alias(c) for c in data.columns]).collect()[0]
        data = data.select([c for c in data.columns if null_percentages[c] < null_percentage])

        # filling missing values by function/value
        if len(features_handling.keys()) > 0:
            missing_values = {
                k: v['fillna'] if not isinstance(v.get('fillna', 'mean'), str) else
                data.agg((eval('F.' + v.get('fillna', 'mean')))(k)).collect()[0][0]
                for (k, v) in features_handling.items()
            }

        # filling default missing features by mean
        default_missing_features = list(set(data.columns).difference(set(list(features_handling.keys()))))
        default_missing_values = data.select([F.mean(c).alias(c) for c in default_missing_features]).collect()[0]
        missing_values.update({c: default_missing_values[c] for c in default_missing_features})
        self.save_metadata('missing_values', missing_values)

        if stored_missing_values is not None:
            data = data.fillna(stored_missing_values)
        else:
            data = data.fillna(missing_values)
        return data

    def _transform_data(self, data):
        data_handling = self.data_settings.get('data_handling', {})

        # interactions
        if data_handling.get('interactions', False):
            columns_list = list(data.columns)
            columns_list.remove(self.model_settings['variable_to_predict'])
            for col1 in columns_list:
                for col2 in columns_list:
                    if col1 != col2:
                        name = str(col1) + '_' + str(col2)
                        reverse_name = str(col2) + '_' + str(col1)
                        if reverse_name not in list(data.columns):
                            data = data.withColumn(name, (F.col(col1) + 1) * (F.col(col2) + 1))

        # binning
        for feature_to_bin in data_handling.get("features_to_bin", []):
            min_val = data.agg({feature_to_bin['name']: "min"}).collect()[0][0]
            max_val = data.agg({feature_to_bin['name']: "max"}).collect()[0][0]
            full_bins = [(min_val - 1)] + feature_to_bin['bins'] + [(max_val + 1)]

            bucketizer = Bucketizer(splits=full_bins,
                                    inputCol=feature_to_bin['name'],
                                    outputCol=feature_to_bin['name'] + '_binned')

            data = bucketizer.transform(data)

        # transformation
        for col in data_handling.get("features_handling", {}).keys():
            transformation_array = data_handling["features_handling"][col].get("transformation", [])
            # applying transformations
            for feature_transformation_method in transformation_array:
                data = data.withColumn(
                    col + '_' + feature_transformation_method, eval('F.' + feature_transformation_method)(col))

        # dropping features
        features_to_remove = data_handling.get('features_to_remove', [])
        if len(features_to_remove) > 0:
            data = data.drop(*[feature for feature in features_to_remove if feature in data.columns])
        return data





from pyspark.ml.feature import Bucketizer
import pyspark.sql.functions as F


def spark_cut(df, col_name, bins, labels):
    """
    Turns a continuous variable into categorical.
    :param df: a spark dataframe
    :param col_name: the continuous column to be categorized.
    :param bins: lower and upper bounds. must be sorted ascending and encompass the col entire range.
    :param labels: labels for each category. should be len(bins)-1
    :return: a spark dataframe with the specified column binned and labeled as specified.
    """
    bucketizer = Bucketizer(splits=bins,
                            inputCol=col_name,
                            outputCol=col_name + '_binned')

    df = bucketizer.transform(df)
    label_array = F.array(*(F.lit(label) for label in labels))
    df = df.withColumn(col_name, label_array.getItem(F.col(col_name + '_binned').cast('integer')))
    df = df.drop(col_name + '_binned')
    return df


def spark_dummies(data, columns=None, drop_first=False):
    """
    returns a new dataframe with a hot vector column for each unique value for each column specified in columns.
    the column itself will be deleted. If no columns are provided, all columns with less then 10 distinct values
    will be converted.
    :param data: a spark dataframe
    :param columns: string or array of strings specifying columns to transform.
    :param drop_first: default=False. If set to True, first column will be removed to avoid multicollinearity.
    :return:  new dataframe with a column for each unique value in the specified columns
    """

    if columns is None:
        columns = [c for c in data.columns if data.select(c).distinct().count() < 10]
    if isinstance(columns, str):
        columns = [columns]
    for col in columns:

        categories = list(map(lambda x: x[0], data.select(col).distinct().collect()))

        dummies = [F.when(F.col(col) == category, 1).otherwise(0).alias(f"{col}_{category}") for
                   category in categories]
        if drop_first:
            dummies.pop(0)
        data = data.select(data.columns + dummies)
        data = data.drop(col)
    return data


def spark_select_dummies(data, column_prefix, targets, combine_to_other=False):
    """
    Use to select dummy variables to keep. values not kept will either be deleted or minimized to a single "other"
    column.
    :param data: a spark dataframe
    :param column_prefix: string. the original column name that was converted to dummies.
    :param targets: String or array of strings. The values to be preserved.
    :param combine_to_other: Default=False. If set to True, columns corresponding to values NOT in target will be
    minimized to a single "colName_other" column.
    :return: new dataframe with non target value columns deleted or minimized.
    """
    if isinstance(targets, str):
        targets = [targets]
    cols_to_combine = []
    for col in data.columns:
        if col.startswith(column_prefix) and col[len(column_prefix)+1:] not in targets:
            if combine_to_other:
                cols_to_combine.append(col)
            else:
                data = data.drop(col)
    if len(cols_to_combine)>0:
        data = data.withColumn(f'{column_prefix}_temp', sum(data[col] for col in cols_to_combine))
        data = data.withColumn(f'{column_prefix}_other', F.when(F.col(f'{column_prefix}_temp') > 0, 1).otherwise(0))
        data = data.drop(f'{column_prefix}_temp')
        data = data.select([c for c in data.columns if c not in cols_to_combine])
    return data



import os
import click
import requests

from mlapp.mlapp_cli.common.cli_utilities import create_directory, create_file
from mlapp.mlapp_cli.common.files import app_file, utilities_file, empty_config_file, default_config_file, \
    env_file, docker_compose_file, run_file, gitignore_file, dockerignore_file
try:
    from mlapp.integrations.aml.cli import _setup_aml as setup_aml_env
except:
    setup_aml_env = None

init_files_directories = ['assets', 'common', 'data', 'deployment', 'env']


def init_command(ml_control_panel, azure_machine_learning, is_gitignore, is_dockerignore, is_force_init):
    if not is_force_init:
        is_initiated = False
        exsiting_files = os.listdir(os.getcwd())
        for f in exsiting_files:
            full_path = os.path.join(os.getcwd(), f)
            if (os.path.isdir(full_path) and f in init_files_directories) or '.py' in f:
                is_initiated = True
                break

        if is_initiated:
            click.secho(
                "ERROR: your project is not empty.\nHint: you can use 'mlapp init --force' option to force init (caution: force may override exsiting files).", fg='red')
            return

    # creates the assets directory if not exists.
    create_directory(directory_name='assets')

    # creates the common directory if not exists.
    create_directory(directory_name='common')

    # creates the data directory if not exists.
    create_directory(directory_name='data', include_init=False)

    # generates app file template
    app_file_content = app_file

    # generates run file template
    run_file_content = run_file

    # generates utilities file template
    utilities_file_content = utilities_file

    config_file_content = empty_config_file
    # create all files from templates
    create_file(file_name='app.py', content=app_file_content)
    create_file(file_name='run.py', content=run_file_content)
    create_file(file_name='utilities.py', path='common', content=utilities_file_content)
    create_file(file_name='config.py', content=config_file_content)
    create_file(file_name='requirements.txt')

    if ml_control_panel:
        # creates the env directory if not exists.
        create_directory(directory_name='env', include_init=False)

        # creates the deployment directory if not exists.
        create_directory(directory_name='deployment', include_init=False)

        docker_compose_file_content = docker_compose_file

        env_file_content = env_file
        # creates files
        create_file(file_name='docker-compose.yaml', path='deployment', content=docker_compose_file_content)
        create_file(file_name='.env', path='env', content=env_file_content)

        # edit content of config.py file, set env new file
        default_env_filename = ""
        config_file_content = default_config_file.replace("<FILENAME>", default_env_filename)
        create_file(file_name='config.py', content=config_file_content)

    if azure_machine_learning:
        if setup_aml_env is not None:
            setup_aml_env()
        else:
            click.secho("Warning: 'azureml sdk is not installed in your environment. please install it and run 'mlapp aml setup' to complete the init operation.", fg='red')

    if not is_gitignore:
        if not os.path.exists(os.path.join(os.getcwd(), '.gitignore')) or is_force_init:
            try:
                # attempt to get gitignore from github
                response = requests.get('https://raw.githubusercontent.com/github/gitignore/master/Python.gitignore')
                if response.status_code!=200:
                    raise RuntimeError('attempted and failed to fetch .gitignore from github.'
                                       ' Will use default')
                # append ml app gitignore
                github_gitignore_file = response.content.decode()
                github_gitignore_file += f'\n{gitignore_file}'

                # dump
                create_file('.gitignore', content=github_gitignore_file)
            except:
                create_file('.gitignore', content=gitignore_file)
        else:
            click.secho("Error: '.gitignore' already exists.", fg='red')

    if not is_dockerignore:
        if not os.path.exists(os.path.join(os.getcwd(), '.dockerignore')) or is_force_init:
            create_file('.dockerignore', content=dockerignore_file)
        else:
            click.secho("Error: '.dockerignore' already exists.", fg='red')



import os
import time
import json
import glob
import datetime
import csv
import logging
import shutil
import sys
import copy
import pandas as pd
import jsonschema
from mlapp.config import settings
from mlapp.managers.pipeline_manager import PipelineManager
from mlapp.utils import general as general_utils
from mlapp.utils.logger import Logger
from mlapp.managers.io_manager import IOManager
from mlapp.handlers.wrappers.database_wrapper import database_instance
from mlapp.handlers.wrappers.file_storage_wrapper import file_storage_instance
from mlapp.handlers.wrappers.spark_wrapper import spark_instance
from mlapp.utils.exceptions.framework_exceptions import SkipToLocalException, UnsupportedFileType, DataFrameNotFound
from mlapp.utils.exceptions.base_exceptions import ConfigKeyError, FrameworkException, ConfigError, JobManagerException
from matplotlib.figure import Figure
try:
    import plotly.graph_objects as go
except ModuleNotFoundError:
    go = None


class JobManager(object):
    def __init__(self, job_id, config, **kwargs):
        """
        Constructor for he job manager.
        Its main role is to populate the input from external environment (e.g. trained model's objects) and
        store the outputs in the relevant location (e.g. store figures, objects in file storage service or in local
        directory in case a file storage is not defined. Store data in database and more.
        The job manager is using the singleton handlers to connect to the different services like DB and FileStore
        :param job_id: the id of this current job.
        :param config: the config string of user configurations for the asset
        :param kwargs: run_id and any initial input to the asset.
        run_id - All outputs files to be created will be given this identifier
        """

        # Handlers:
        self.output_logger_filename = "output_logs.csv"
        self.file_store_handler = file_storage_instance
        self.db_handler = database_instance
        self.spark_handler = spark_instance

        self.file_objects_types = {
            "pyspark": {
                "type": ("pyspark" if self.file_store_handler.empty() else "pyspark.zip"),
                "load_method": self._load_files_into_spark_object,
                "store_method": self._store_spark_object_into_files
            },
            "pkl": {
                "type": "pkl.pkl",
                "load_method": self._load_pickle_as_object,
                "store_method": self._store_object_as_pickle
            }
            # "tensorflow",
            # "pytorch",
            # "keras",
        }

        # config parts
        self.job_id = job_id

        self.config = config
        self.data_settings = config.get('data_settings', {})
        self.model_settings = config.get('model_settings', {})
        self.flow_settings = config.get('flow_settings', {})
        self.job_settings = config.get('job_settings', {})

        self.filestore_buckets = settings.get("file_store_buckets", {})
        self.local_storage_path = settings.get("local_storage_path", "output")
        self.temporary_storage_path = settings.get("temporary_storage_path", "temporary_output")
        self.temporary_storage_files = set()
        self.set_output_folders()
        self.start_time = time.strftime('%Y-%m-%d %H:%M:%S')
        self.last_time = self.start_time
        self.deploy_environment = settings.get("deploy_environment", "default")
        self.deploy_version = settings.get("deploy_version", "-")

        self.identity = {
            'run_id': kwargs['run_id'],
            'pipeline': self.job_settings.get('pipeline', None),
            'asset_name': self.get_asset_name(kwargs.get('has_flow_summary')),
            'asset_label': self.job_settings.get('asset_label', None)
        }

        # consts
        self.DOT = '.'
        self.LOAD = 'LOAD'
        self.SAVE = 'SAVE'
        self.DATA = 'data'
        self.MODELS = 'models'
        self.FEATURES = 'features'
        self.INPUT_FROM_PREDECESSOR = 'input_from_predecessor'

        self.input_manager = IOManager()
        self.output_manager = IOManager()

    def __del__(self):
        self.clean_output_folders()

    ##########################################################################
    #                                                                        #
    #                            Run Pipeline                                #
    #                                                                        #
    ##########################################################################
    def run_pipeline(self, *args, **kwargs):
        """
        Loads the inputs defined in the config ( e.g. pre trained objects, metadata and more)
        Initiates a pipeline manager, runs it
        Stores the outputs in the relevant locations
        :param args: arguments to be passed to the first stage in the pipeline
        :param kwargs: keyword arguments to be loaded into the Input (an IOManager instance)
        :return: IOManager instance with all the outputs to be stored
        """
        try:
            self._store_job_config()
            self.log_to_file(run_id=None)  # no `run_id` output yet
            self.validate_config()
            self.load_input(**kwargs)
            pipeline_manager = PipelineManager(
                self.identity['run_id'],
                self.job_settings['pipeline'],
                _input=self.input_manager, _output=self.output_manager,
                config=copy.deepcopy(self.config),
                **kwargs
            )
            self.pipeline_start_print(self.identity['run_id'])
            pipeline_manager.run(*args)
            self.pipeline_end_print(self.identity['run_id'])
            self.store_output()
            self.temporary_files_message()
        except Exception as error:
            self.identity['run_id'] = None    # pipeline failed
            log = logging.getLogger(self.job_id)
            log.error(str(error), exc_info=True)
            raise error
        finally:
            # saving logger file
            self._store_logger_file(run_id=self.identity['run_id'])
        return self.output_manager

    ##########################################################################
    #                                                                        #
    #                             Get Functions                              #
    #                                                                        #
    ##########################################################################

    def get_custom_filename(self, run_id, name):
        """
        :param run_id: unique id of the asset
        :param name: file name
        :return: proper concatenation of the file name
        """
        return str(run_id) + '_' + self.identity['asset_name'] + '_' + name

    def get_dataframe_filename(self, run_id, df_name):
        """
        :param run_id: unique id of the asset
        :param df_name: dataframe name
        :return: proper concatenation of the dataframe file name
        """
        return str(run_id) + '_' + self.identity['asset_name'] + self.DOT + df_name + '.csv'

    def get_config_filename(self, run_id):
        """
        :param run_id: unique id of the asset
        :return: proper concatenation of the config file name
        """
        return str(run_id) + '_' + self.identity['asset_name'] + '.config.json'

    def get_job_config_filename(self):
        """
        :return: proper concatenation of the config file name
        """
        return str(self.job_id) + '.config.json'

    def get_metadata_filename(self, run_id):
        """
        :param run_id: unique id of the asset
        :return: proper concatenation of the metadata file name
        """
        return str(run_id) + '_' + self.identity['asset_name'] + '.metadata.json'

    @staticmethod
    def get_flow_metadata_filename(run_id):
        """
        :param run_id:  unique id of the asset
        :return: proper concatenation of the flow-summary-asset metadata file name
        """
        return str(run_id) + '_flow_summary.metadata.json'

    def get_features_filename(self, run_id):
        """
        :param run_id: unique id of the asset
        :return: proper concatenation of the feature file name
        """
        return str(run_id) + '_' + self.identity['asset_name'] + '.features.csv'

    def get_objects_filename(self, run_id, manager, mlapp_type, file_type, model_name=None, class_name=None):
        """
        :param run_id: string, unique id of the asset
        :param manager: string , manager name
        :param mlapp_type: string , one of the keys in self.file_objects_types
        :param file_type: string , file type for example, .csv .txt etc
        :param model_name: string , name of the asset
        :param class_name: string, spark object class name.
        :return: proper concatenation of the object file name
        """
        if model_name is None and class_name is None:
            return str(run_id) + '_' + self.identity['asset_name'] + \
                   self.DOT + manager + self.DOT + mlapp_type + self.DOT + file_type
        elif class_name is None:
            return str(run_id) + '_' + self.identity['asset_name'] + self.DOT + model_name + \
                   self.DOT + manager + self.DOT + mlapp_type + self.DOT + file_type
        elif model_name is None:
            return str(run_id) + '_' + self.identity['asset_name'] + self.DOT + class_name + \
                   self.DOT + manager + self.DOT + mlapp_type + self.DOT + file_type
        else:
            return str(run_id) + '_' + self.identity['asset_name'] + self.DOT + class_name + \
                   self.DOT + model_name + self.DOT + manager + self.DOT + mlapp_type + self.DOT + file_type

    def get_objects_modules_filename(self, run_id):
        """
        :param run_id: string, unique id of the asset
        :return: proper concatenation of the object modules json file name
        """
        return str(run_id) + '_' + self.identity['asset_name'] + '.objects.modules.json'

    def get_logger_filename(self, run_id=None):
        """
        :param run_id: string, unique id of the asset
        :return: proper concatenation of the logger file name
        """
        if run_id is not None:
            return str(run_id) + '_' + str(self.job_id) + '.logger.txt'
        else:
            return str(self.job_id) + '.logger.txt'

    def get_asset_name(self, has_flow_summary):
        """
        :param has_flow_summary: if the config contains flow summary section
        :return: model name from the config
        """
        if has_flow_summary is not None and not has_flow_summary:
            return None
        else:
            asset_name = self.job_settings.get('asset_name', None)
            if asset_name is None:
                #  model_name deprecation
                asset_name = self.job_settings.get('model_name', None)
                if asset_name is None:
                    raise ConfigKeyError('asset_name is required in job_settings config.')
                else:
                    try:
                        raise DeprecationWarning('model_name is deprecated please use asset_name instead.')
                    except Warning as w:
                        print("DeprecationWarning: " + str(w))
            return asset_name

    ###############################################################
    #                                                             #
    #                        Load Input                           #
    #                                                             #
    ###############################################################
    def load_input(self, **kwargs):
        """
          This is Main Load Input function that called in run_pipeline function.
          It loads any input specified in the config (e.g. trained model object, saved features DataFrame etc.).
          The function will load the input into an IOManager instance - the InputManager.
          :param kwargs:  input from predecessor pipeline. Passed by the flow manager.
          :return: None

        :param kwargs:
        :return:
        """

        # Todo: 1. merge in kwargs into the input_manager by calling the update_recursive_dict(input,kwargs)
        if kwargs:
            self.input_manager.add_analysis_metadata(self.INPUT_FROM_PREDECESSOR, kwargs)

        # get ids from config
        config_model_id = self._load_model_id_from_config()
        config_data_id = self._load_data_id_from_config()
        config_reuse_features_id = self._load_reuse_features_id_from_config()
        # config_reuse_features_flag = self._load_reuse_features_flag_from_config()

        # if there is no ids in the config, skip load input method.
        if config_model_id is None and config_data_id is None and config_reuse_features_id is None:
            return

        # loads required files from storage.
        model_config = self.load_config(config_model_id)
        # looking for the right config_data_id
        config_data_id = self._decide_on_config_data_id(config_data_id, model_config, config_model_id)
        data_config = self.load_config(config_data_id)
        train_metadata = self.load_metadata(config_model_id)
        train_features = self.load_features(config_reuse_features_id)
        train_objects = self.load_objects(config_model_id, config_data_id)

        # updates config if necessary
        self.update_config(model_config, data_config)

        # gets train data and models metadata
        if train_metadata is not None:
            train_data_metadata = train_metadata.get(self.DATA, {})
            train_models_metadata = train_metadata.get(self.MODELS, {})

            # add analysis metadata to input manager
            self.input_manager.add_analysis_metadata(self.DATA, train_data_metadata)
            self.input_manager.add_analysis_metadata(self.MODELS, train_models_metadata)

        # gets train data and models objects
        if train_objects is not None:
            train_data_objects = train_objects.get(self.DATA, {})
            train_models_objects = train_objects.get(self.MODELS, {})

            # add objects to input manager
            self.input_manager.add_objects(self.DATA, train_data_objects)
            self.input_manager.add_objects(self.MODELS, train_models_objects)

        # check if features reused
        if train_features is not None:
            self.input_manager.add_dataframe(self.FEATURES, train_features)

    @staticmethod
    def _decide_on_config_data_id(config_data_id, model_config, config_model_id):
        """
        Decide on the right data_id for model forecast to reuse
        scenario 1: data_id is supplied in the main config
        scenario 2: data_id missing in main config, but supplied in the reused model config (the train model config)
        scenario 3: data_id is missing and is missing in the reused model config, in this case data_id = model_id
        @param config_data_id: string
        @param model_config: dict
        @param config_model_id: string
        @return: string
        """
        if config_data_id is None:
            if model_config is None:
                return
            config_model_data_id = model_config.get('task_settings', {}).get('data_id', None)
            if config_model_data_id is None:
                new_config_data_id = config_model_id
            else:
                new_config_data_id = config_model_data_id
        else:
            new_config_data_id = config_data_id
        return new_config_data_id

    def load_config(self, config_run_id):
        """
        loads the config file of a previous asset run from filesystem
        :param config_run_id: the id of the run
        :return:
        """
        if config_run_id is None:
            return
        config_filename = self.get_config_filename(config_run_id)
        config = self._load_json_as_object(config_filename, bucket_name='configs')
        return config

    def load_metadata(self, run_id):
        """
        loads the metadata file of a previous asset run from filesystem
        :param run_id: the id of the run
        :return:
        """
        if run_id is None:
            return
        data_output_filename = self.get_metadata_filename(run_id)
        data_results = self._load_json_as_object(data_output_filename, bucket_name='metadata')
        return data_results

    def load_features_with_flag(self, data_id, reuse_flag=False):
        """
        :param data_id:
        :param reuse_flag:
        :return:
        """
        if not reuse_flag or data_id is None:
            return
        features_filename = self.get_features_filename(data_id)
        features = self._load_csv_as_data_frame(features_filename, bucket_name='csvs')
        return features

    def load_features(self, config_reuse_features_id):
        """
        loads the features file of a previous asset run from filesystem
        :param config_reuse_features_id: the id of the run that created the features
        :return:
        """
        if config_reuse_features_id is None:
            return
        features_filename = self.get_features_filename(config_reuse_features_id)
        features = self._load_csv_as_data_frame(features_filename, bucket_name='csvs')
        return features

    def load_objects(self, model_id, data_id):
        """
        loads the object files of a previous asset run from filesystem
        :param model_id: run_id id of the run that created the model
        :param data_id: run_id id of the run that created the features
        :return: dictionary with objects
        """
        if model_id is None and data_id is None:
            return
        object_results_filenames = self._load_model_objects_files_names(model_id, data_id)
        object_results = {}
        try:
            modules = self._load_objects_modules_file(model_id)
            for f in object_results_filenames:
                nargs = len(f.split('.'))
                if nargs == 4:
                    file_name, manager, mlapp_type, file_type = f.split('.')
                    object_results[manager] = \
                        self.file_objects_types.get(mlapp_type, self.file_objects_types['pkl'])['load_method'](
                            f, bucket_name='objects')
                else:
                    # first 6 types of names
                    file_name, class_name, model_name, manager, mlapp_type, file_type = f.split('.')[:6]
                    import_module = self.import_object_from_module(module=modules.get(class_name, None),
                                                                   class_name=class_name)
                    if manager not in object_results.keys():
                        object_results[manager] = {}
                    object_results[manager][model_name] = \
                        self.file_objects_types.get(mlapp_type, self.file_objects_types['pkl'])['load_method'](
                            f, bucket_name='objects', module=import_module)

            return object_results
        except Exception as e:
            print("An error accord while load objects. Reason: " + str(e))
            raise FrameworkException()

    ##################################################################
    #                                                                #
    #                       Load Private Functions                   #
    #                                                                #
    ##################################################################

    def _load_model_id_from_config(self):
        model_id = self.job_settings.get('model_id', None)
        return model_id

    def _load_data_id_from_config(self):
        data_id = self.job_settings.get('data_id', None)
        return data_id

    def _load_reuse_features_id_from_config(self):
        reuse_features_id = self.job_settings.get('reuse_features_id', None)
        return reuse_features_id

    def _load_reuse_features_flag_from_config(self):
        reuse_features_flag = self.job_settings.get('reuse_features', False)
        return reuse_features_flag

    def _load_pickle_as_object(self, file_name, module=None, bucket_name='objects'):
        try:
            if self.file_store_handler.empty() or self.filestore_buckets.get(bucket_name) is None:
                raise SkipToLocalException()

            local_path = self._load_file_from_file_store(self.filestore_buckets[bucket_name], file_name)
        except SkipToLocalException:
            local_path = os.path.join(self.local_storage_path, file_name)
        except Exception as e:
            print("Failed to load object from file storage. Reason: " + str(e))
            raise FrameworkException()

        return general_utils.load_pickle_to_object(local_path)

    def _load_files_into_spark_object(self, file_name, module=None, bucket_name='objects'):
        try:
            if module is None:
                raise JobManagerException('spark module object path must be passed in order to load a spark model.')

            if self.file_store_handler.empty():
                raise SkipToLocalException()

            file_path = self._load_file_from_file_store(self.filestore_buckets[bucket_name], file_name)
            extract_dir = os.path.splitext(file_name)[0]  # rm the word zip
            if self.file_store_handler.empty():
                general_utils.uncompress_zip_to_folder(
                    file_path, extract_dir=os.path.join(self.local_storage_path, extract_dir))
            else:
                general_utils.uncompress_zip_to_folder(
                    file_path, extract_dir=os.path.join(self.temporary_storage_path, extract_dir))
        except SkipToLocalException:
            try:
                file_path = os.path.join(self.local_storage_path, file_name)
            except Exception as e:
                print("spark model file was not loaded. Reason: " + str(e))
                return None
        except Exception as e:
            print("Failed to load spark object from file storage. Reason: " + str(e))
            raise FrameworkException()

        return self.spark_handler.load_model(file_path=file_path.replace('.zip', ''), module=module)

    def _load_objects_modules_file(self, model_id):
        try:
            if model_id is None:
                return
            modules_filename = self.get_objects_modules_filename(model_id)
            modules = self._load_json_as_object(modules_filename, bucket_name='objects')
            return modules
        except:
            return {}

    def _load_model_objects_files_names(self, model_id, data_id):
        files_path = []
        try:
            types_suffix = list(map(lambda s: s.get('type', 'pkl'), self.file_objects_types.values()))

            if self.file_store_handler.empty():
                raise SkipToLocalException()

            if model_id is not None:
                files_path.extend(map(lambda x: x.decode("utf-8") if hasattr(x, 'decode') else x,
                                      self._load_files_names_from_file_storage(model_id)))

            if data_id is not None:
                files_path.extend(map(lambda x: x.decode("utf-8") if hasattr(x, 'decode') else x,
                                      self._load_files_names_from_file_storage(data_id)))

            files_path = filter(lambda f: f.endswith(tuple(types_suffix)), files_path)

        except SkipToLocalException:
            if model_id is not None:
                files_path.extend(self._load_files_names_from_local_storage(model_id))

            if data_id is not None:
                files_path.extend(self._load_files_names_from_local_storage(data_id))

            files_path = filter(lambda f: f.endswith(tuple(types_suffix)), files_path)
        except Exception as e:
            raise FrameworkException('Failed to load objects files from file storage. reason: ' + str(e))

        return files_path

    def _load_files_names_from_local_storage(self, run_id):
        # reads all files starting with id as prefix local storage.
        return list(map(lambda f: os.path.basename(f),
                        glob.glob(os.path.join(self.local_storage_path, run_id + "_*"))))

    def _load_files_names_from_file_storage(self, run_id):
        # reads all files starting with id as prefix from file storage.
        return self.file_store_handler.list_files(
            bucket_name=self.filestore_buckets.get('objects'), prefix=run_id, **self.identity)

    def _load_csv_as_data_frame(self, file_name, bucket_name='csvs'):
        try:
            if self.file_store_handler.empty() or self.filestore_buckets.get(bucket_name) is None:
                raise SkipToLocalException()

            local_file_path = self._load_file_from_file_store(self.filestore_buckets[bucket_name], file_name)
            df = pd.read_csv(local_file_path)

        except SkipToLocalException:
            try:
                df = pd.read_csv(os.path.join(self.local_storage_path, file_name))
            except Exception as e:
                print("File %s was not loaded. Reason: %s" % (str(file_name), str(e)))
                raise DataFrameNotFound(
                    "Error: file " + file_name + " is not found. Make sure you provided the right names and ids.")
        except Exception as e:
            print("File %s was not loaded. Reason: %s" % (str(file_name), str(e)))
            raise FrameworkException()
        return df

    def _load_json_as_object(self, file_name, bucket_name=None):
        try:
            if self.file_store_handler.empty() or self.filestore_buckets.get(bucket_name) is None:
                raise SkipToLocalException()

            local_path = self._load_file_from_file_store(self.filestore_buckets[bucket_name], file_name)
        except SkipToLocalException:
            local_path = os.path.join(self.local_storage_path, file_name)
        except Exception as e:
            raise FrameworkException(str(e))
        return general_utils.read_json_file(local_path)

    def _load_file_from_file_store(self, bucket_name, filename):
        unicode_file_name = str(filename)
        try:
            if self.file_store_handler.empty():
                raise SkipToLocalException()
            try:
                # trying to get file from temporary_storage_path
                to_path = os.path.join(self.temporary_storage_path, unicode_file_name)
                self.file_store_handler.download_file(bucket_name, unicode_file_name, to_path, **self.identity)
            except Exception:
                try:
                    # trying to get file from local_storage_path
                    to_path = os.path.join(self.local_storage_path, unicode_file_name)
                    self.file_store_handler.download_file(bucket_name, unicode_file_name, to_path, **self.identity)
                except Exception as e:
                    raise FrameworkException(str(e))
            return to_path
        except SkipToLocalException as e:
            raise FrameworkException(str(e))

    ###############################################################
    #                                                             #
    #                Store Output Main Functions                  #
    #                                                             #
    ###############################################################

    def store_output(self):
        """
        This is Main store output function that is called in run_pipeline function
        It goes through the different structures in OutputManager of the asset and stores them
        :return: None
        """
        # call store methods
        self.store_configs()
        self.store_metadata()
        self.store_dataframes()
        self.store_objects()
        self.store_images()

    def store_dataframes(self):
        """
        Stores the DataFrame/s saved by the data scientist in the outputManager
        Iterates over dataframes and stores each DataFrame according to its key
        :return: None
        """
        # getting dataframes from output manager
        dataframes = self.output_manager.get_dataframes()
        tables = self.output_manager.get_tables()

        # iterate over dataframes and storing each dataframe according to key
        # if to table is not None or empty string, it will try to store the df in DB.
        for df_name, df in dataframes.items():
            if df is None:
                print("Warning: %s Dataframe not stored" % (str(df_name)))
            table_name = tables.get(df_name)
            filename = self.get_dataframe_filename(self.identity['run_id'], df_name)
            self._store_dataframe(df, filename, run_id=self.identity['run_id'], table_name=table_name)

    def store_objects(self):
        """
         Stores the objects saved by the data scientist in the outputManager
         Iterates over file types saved in output manager and store each object according to its type
        :return: None
        """
        # getting object from output manager
        objects = self.output_manager.get_objects()

        # iterate over file types saved in output manager and storing each object according to its type
        for file_type, content in objects.items():

            # gets mlapp relevant type dictionary
            obj_dict = self.file_objects_types.get(file_type, self.file_objects_types['pkl'])

            # handle model manager storing objects
            models_content = content.get(self.MODELS, {})
            if models_content:
                # gets mlapp_file_type
                try:
                    mlapp_type, _ = obj_dict['type'].split(self.DOT)
                except:
                    mlapp_type = obj_dict['type']

                if isinstance(models_content, dict) and mlapp_type == 'pyspark':
                    objects_modules = {}
                    for model_name, model_object in models_content.items():
                        # gets models file name according to file type
                        model_filename = self.get_objects_filename(self.identity['run_id'], "models", mlapp_type,
                                                                   file_type, model_name=model_name,
                                                                   class_name=model_object.__class__.__name__)
                        # call store method according to file type
                        obj_dict['store_method'](model_object, model_filename, bucket_name='objects')

                        # save object module for loading
                        objects_modules[model_object.__class__.__name__] = model_object.__module__

                    # stores objects modules in a json file
                    self._store_object_modules(objects_modules, bucket_name='objects')

                else:
                    # gets models file name according to file type
                    models_filename = self.get_objects_filename(
                        self.identity['run_id'], "models", mlapp_type, file_type, model_name=None)

                    # call store method according to file type
                    obj_dict['store_method'](models_content, models_filename, bucket_name='objects')

            # handle data manager storing objects
            data_content = content.get(self.DATA, {})
            if data_content:
                # gets mlapp_type
                mlapp_type, _ = obj_dict['type'].split(self.DOT)

                # gets data file name according to file type
                data_filename = self.get_objects_filename(self.identity['run_id'], "data", mlapp_type, file_type)

                # call store method according to file type
                obj_dict['store_method'](data_content, data_filename, bucket_name='objects')

    def store_metadata(self):
        """
        This function stores the model metadata either in the database, or in a file storage as a JSON,
        or in the CSV logger file
        :return: None
        """
        run_id = self.identity['run_id']
        file_name = self.get_metadata_filename(run_id)
        metadata = self.output_manager.get_metadata()
        json_validation_warning_message = "%s is not json serializable, removed from metadata."
        general_utils.validate_json(metadata, json_validation_warning_message)

        try:
            if self.db_handler.empty():
                raise SkipToLocalException("No database set up in 'config.py'.")

            # trying to save metadata in db
            self._store_metadata_in_db(run_id, metadata=metadata)
            print("Stored the metadata in the db.")
        except SkipToLocalException:
            pass
        except Exception as e:
            print("Failed to store metadata in db. Reason: " + str(e))

        saved_path = None
        try:
            saved_path = self._store_object_as_json(metadata, file_name, bucket_name='metadata')
        except Exception as e:
            print("Error: in storing metadata please check if your metadata is a valid JSON. Reason: " + str(e))

        try:
            if os.path.exists(saved_path):
                self._store_object_in_csv_logger(run_id, metadata)
            print("Saved metadata in the local CSV logger file.")
        except Exception as e:
            print("Failed to store metadata in CSV logger file. Reason: " + str(e))

    def store_flow_metadata(self, flow_id, **kwargs):
        """
        This function stores the flow_job_manager metadata and relevant only for the flow summary job manager.
        It stores the flow metadata either in the database, or in a file storage as a JSON, or in the CSV logger file
        :param flow_id: string
        :return: None
        """
        metadata = {}
        pipelines_metadata = kwargs.get('pipelines_metadata', {})
        metadata['flow_id'] = flow_id
        metadata['pipelines_metadata'] = pipelines_metadata
        json_validation_warninig_message = "%s is not json serializable, removed from metadata."
        general_utils.validate_json(metadata, json_validation_warninig_message)
        file_name = self.get_flow_metadata_filename(flow_id)

        # trying to save analysis results in db
        try:
            if self.db_handler.empty():
                raise SkipToLocalException("No database set up in 'config.py'.")

            self._store_flow_metadata_in_db(flow_id, metadata=metadata, properties={})
            print("Stored the metadata in the db.")
        except SkipToLocalException:
            pass
        except Exception as e:
            print("Failed to store metadata in db. Reason: " + str(e))

        saved_path = None
        try:
            saved_path = self._store_object_as_json(metadata, file_name, bucket_name='metadata')
        except Exception as e:
            print("Error: in storing metadata please check if your metadata is a valid JSON. Reason: " + str(e))

        try:
            if os.path.exists(saved_path):
                self._store_object_in_csv_logger(flow_id, metadata)
                print("Stored the metadata in the local CSV logger file.")
        except Exception as e:
            print("Failed to store metadata in CSV logger file. Reason: " + str(e))

    def store_configs(self):
        """
        Stores the config of this mlapp run as json file. Saves in Filestore if defined or locally.
        :return: None
        """
        model_config = self.config.copy()
        file_name = self.get_config_filename(self.identity['run_id'])
        self._store_object_as_json(model_config, file_name, bucket_name='configs')

    def store_images(self, bucket_name='imgs'):
        """
        Stores the images saved by the data scientist in the outputManager. Saves in Filestore if defined or locally.
        :param bucket_name: dedicated bucket name in the FileStore
        :return: None
        """
        images = self.output_manager.get_images_files()
        images_for_upload = []

        # saving images in local storage and preparing them for upload
        model_filename = self.get_custom_filename(self.identity['run_id'], '')
        for image_name, figure in images.items():
            if go and isinstance(figure, go.Figure):
                file_name = model_filename + image_name + '.html'
            elif isinstance(figure, Figure):
                file_name = model_filename + image_name + '.png'
            else:
                raise UnsupportedFileType(
                    f"{image_name} is not in the right format. "
                    f"ML App supports only figures from the matplotlib or plotly.js library.")

            if self.file_store_handler.empty():
                file_path = os.path.join(self.local_storage_path, file_name)
            else:
                file_path = os.path.join(self.temporary_storage_path, file_name)
            try:
                if go and isinstance(figure, go.Figure):
                    figure.write_html(file_path, include_plotlyjs='directory')
                else:
                    general_utils.save_figure_to_png(figure, file_path, file_name)
                images_for_upload.append({
                    'file_name': file_name,
                    'file_path': file_path
                })
            except Exception as e:
                print(e)
        # saving images in file storage
        if images_for_upload:
            try:
                if self.file_store_handler.empty() or self.filestore_buckets.get(bucket_name) is None:
                    raise SkipToLocalException("No file storage set up in 'config.py'")
                for image in images_for_upload:
                    self._store_file_in_file_store(self.filestore_buckets['imgs'], image['file_name'],
                                                   image['file_path'])
                    os.remove(image['file_path'])
            except SkipToLocalException:
                pass
            except Exception as e:
                # Keeping images in local file store in case of failure
                print("Failed to store images in file storage. Reason: " + str(e))
                print("Stored images in local file storage instead.")

    ###################################################################
    #                                                                 #
    #                       Store Private Functions                   #
    #                                                                 #
    ###################################################################

    def _store_job_config(self):
        job_config = self.config.copy()
        file_name = self.get_job_config_filename()
        self._store_object_as_json(job_config, file_name, bucket_name='configs')

    def _store_metadata_in_db(self, run_id, metadata, properties=None, table='analysis_results'):
        """
        Storing meta in db
        :param run_id: id of the current running pipeline
        :param metadata: analysis results object to store in db
        :return: None
        """
        if properties is None:
            properties = {}

        model_properties = {'job_id': self.job_id, 'deploy_version': self.deploy_version}
        if self.job_settings.get('asset_label', False):
            model_properties['asset_label'] = self.job_settings['asset_label']
        if self.job_settings.get('batch_name', False):
            model_properties['batch_name'] = self.job_settings['batch_name']
        if properties:
            for prop, content in properties.items():
                model_properties[prop] = content

        config_str = json.dumps(model_properties)
        metadata = json.dumps(metadata, cls=general_utils.NumpyEncoder)

        if isinstance(self.identity['pipeline'], str):
            curr_pipeline = self.identity['pipeline']
        else:
            curr_pipeline = ','.join(self.identity['pipeline'])

        data_to_insert = {
            'model_id': run_id,
            'asset_name': self.identity['asset_name'],
            'pipeline': curr_pipeline,
            'properties': config_str,
            'metadata': metadata,
            'environment': self.deploy_environment,
            'created_at': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        }
        if self.job_settings.get('asset_label', False):
            data_to_insert['asset_label'] = self.job_settings['asset_label']

        df_for_insert = pd.DataFrame(data=[data_to_insert])

        self.db_handler.insert_df(table, df_for_insert)

    def _store_flow_metadata_in_db(self, flow_id, metadata, properties=None, table='flows'):
        """
        Storing meta in db
        :param flow_id: id of the current running model
        :param metadata: analysis results object to store in db
        :return: None
        """
        if properties is None:
            properties = {}

        model_properties = {'job_id': self.job_id}
        if self.job_settings.get('asset_label', False):
            model_properties['asset_label'] = self.job_settings['asset_label']
        if self.job_settings.get('batch_name', False):
            model_properties['batch_name'] = self.job_settings['batch_name']
        if properties:
            for prop, content in properties.items():
                model_properties[prop] = content

        config_str = json.dumps(model_properties)
        metadata = json.dumps(metadata, cls=general_utils.NumpyEncoder)

        df_for_insert = pd.DataFrame(data=[{
            'flow_id': flow_id,
            'metadata': metadata,
            'properties': config_str,
            'created_at': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        }])

        self.db_handler.insert_df(table, df_for_insert)

    def _store_logger_file(self, run_id=None, bucket_name='logs'):
        log = logging.getLogger(self.job_id)
        for handler in log.handlers:
            handler.close()
            log.removeHandler(handler)
        sys.stdout = sys.__stdout__

        try:
            if self.file_store_handler.empty():
                raise SkipToLocalException("No file storage set up in 'config.py'")
            logger_file_name = self.get_logger_filename()
            fallback_path = os.path.join(self.temporary_storage_path, logger_file_name)
            if os.path.exists(fallback_path):
                if run_id is not None:
                    logger_file_name = self.get_logger_filename(run_id)
                    new_fallback_path = os.path.join(self.temporary_storage_path, logger_file_name)
                    shutil.move(fallback_path, new_fallback_path)
                    fallback_path = new_fallback_path
                try:
                    self._store_file_in_file_store(
                        self.filestore_buckets[bucket_name], logger_file_name, fallback_path)
                    os.remove(fallback_path)
                    print("Stored logger file %s in the file storage." % str(logger_file_name))
                except Exception as e:
                    print("Failed to store logger file %s in the file storage. Reason: %s" %
                          (str(logger_file_name), str(e)))
                    print("Logger file saved in %s storage path instead." % str(self.temporary_storage_path))
                    self.temporary_storage_files.add(str(logger_file_name))
        except SkipToLocalException:
            logger_file_name = self.get_logger_filename()
            local_path = os.path.join(self.local_storage_path, logger_file_name)
            if os.path.exists(local_path):

                if run_id is not None:
                    logger_file_name = self.get_logger_filename(run_id)
                    new_local_path = os.path.join(self.local_storage_path, logger_file_name)
                    os.rename(local_path, new_local_path)
                    print("Stored logger file %s in the local storage." % str(logger_file_name))
        except Exception as e:
            print('An Error accord when trying to write in to logger file. Reason: %s' % str(e))

    def _store_object_as_pickle(self, obj, file_name, bucket_name='objects'):
        # Trying to save model pickle in file store
        try:
            if self.file_store_handler.empty():
                raise SkipToLocalException("No file storage set up in 'config.py'")

            file_path = os.path.join(self.temporary_storage_path, file_name)
            general_utils.save_object_to_pickle(obj, file_path)
            self._store_file_in_file_store(self.filestore_buckets[bucket_name], file_name, file_path)
            os.remove(file_path)
        except SkipToLocalException:
            file_path = os.path.join(self.local_storage_path, file_name)
            general_utils.save_object_to_pickle(obj, file_path)
            print("Stored model in file %s in local storage %s" % (str(file_name), str(self.local_storage_path)))
        except Exception as e:
            print("Failed to store model pickle in file storage. Reason: " + str(e))
            print("Model is saved in %s storage path instead." % str(self.temporary_storage_path))
            self.temporary_storage_files.add(str(file_name))

    def _store_file_in_file_store(self, bucket_name, file_name, file_local_path):
        self.file_store_handler.upload_file(bucket_name, file_name, file_local_path, **self.identity)

    def _store_object_as_json(self, obj, file_name, bucket_name=None):
        saved_path = os.path.join(self.temporary_storage_path, file_name)
        try:
            if self.file_store_handler.empty():
                raise SkipToLocalException("No file storage set up in 'config.py'")

            # saving object as json
            general_utils.save_object_as_json(obj, saved_path)

            if self.job_settings.get('compress_files', False):
                general_utils.get_compressed_file_and_remove_uncompressed(saved_path)
                saved_path += ".zip"
                file_name += ".zip"

            # uploading to file store
            self._store_file_in_file_store(self.filestore_buckets[bucket_name], file_name, saved_path)
            os.remove(saved_path)
        except SkipToLocalException:
            # saving locally in local_storage_path
            try:
                saved_path = os.path.join(self.local_storage_path, file_name)

                # saving object as json
                general_utils.save_object_as_json(obj, saved_path)
            except Exception as e:
                print("Failed to store json %s in local storage. Reason: %s" % (str(file_name), str(e)))
                raise FrameworkException()
        except Exception as e:
            print("Failed to store json %s in file storage. Reason: %s" % (str(file_name), str(e)))
            if os.path.exists(saved_path):
                self.temporary_storage_files.add(str(file_name))
                print('Stored json %s in %s' % (str(file_name), str(self.temporary_storage_path)))
                self.temporary_storage_files.add(str(file_name))

        return saved_path

    def _store_object_in_csv_logger(self, run_id, dictionary):
        """
        Storing analysis results in csv logger
        :param run_id: current running run_id id
        :param dictionary: static dictionary
        :return: None
        """
        row = [key + ": " + str(value) for key, value in dictionary.items()]
        row.insert(0, "run_id: " + str(run_id))
        file_path = os.path.join(self.local_storage_path, self.output_logger_filename)
        with open(file_path, 'a+') as f:
            writer = csv.writer(f)
            writer.writerow(row)

    def _store_dataframe(self, df, file_name, run_id=None, dtype=None, bucket_name='csvs', table_name=None):
        # trying to insert data frame to the database
        try:
            if self.db_handler.empty() or table_name is None:
                raise SkipToLocalException("No Database set up in 'config.py'.")

            self._store_dataframe_in_db(df, table_name, run_id=run_id, dtype=dtype)
        except Exception as e:
            if not isinstance(e, SkipToLocalException):
                print("Failed to store Dataframe in database table %s. Reason: %s" % (str(table_name), str(e)))

            self._store_data_frame_as_csv(df, file_name, bucket_name=bucket_name)

    def _store_dataframe_in_db(self, df, table_name, run_id=None, dtype=None):
        config_model_id = self._load_model_id_from_config()

        # TODO: split index into columns and add asset name and asset label to table name
        # adding column of analysis id
        if run_id is not None:
            df['forecast_id'] = run_id

            if config_model_id is None:
                df['model_id'] = run_id
            else:
                df['model_id'] = config_model_id

            df['timestamp'] = pd.datetime.now()

        self.db_handler.insert_df(table_name, df)  # TODO: add dtypes to db handlers

    def _store_data_frame_as_csv(self, df, file_name, bucket_name='csvs'):
        # trying to save predictions in file store
        try:
            if self.file_store_handler.empty():
                raise SkipToLocalException("No file storage set up in 'config.py'.")

            file_path = os.path.join(self.temporary_storage_path, file_name)

            # saving to backup path
            df.to_csv(file_path, index=False)

            self._store_file_in_file_store(self.filestore_buckets[bucket_name], file_name, file_path)
            os.remove(file_path)
        except SkipToLocalException:
            file_path = os.path.join(self.local_storage_path, file_name)
            # saving to local path
            df.to_csv(file_path, index=False)
            print("Stored %s in local storage %s" % (str(file_name), str(self.local_storage_path)))
        except Exception as e:
            print("Failed to store Dataframe file %s. Reason: %s" % (str(file_name), str(e)))
            print("Stored %s in %s" % (str(file_name), str(self.temporary_storage_path)))
            self.temporary_storage_files.add(str(file_name))

    def _store_object_modules(self, data, bucket_name=None):
        file_name = self.get_objects_modules_filename(self.identity['run_id'])
        try:
            if self.file_store_handler.empty():
                raise SkipToLocalException("No file storage set up in 'config.py'")

            file_path = os.path.join(self.temporary_storage_path, file_name)

            if not os.path.exists(file_path):
                general_utils.save_object_as_json(data, file_path)
            else:
                general_utils.update_json_file(data, file_path)

            self._store_file_in_file_store(self.filestore_buckets[bucket_name], file_name, file_path)
            os.remove(file_path)
        except SkipToLocalException:
            # storing_local
            file_path = os.path.join(self.local_storage_path, file_name)

            if not os.path.exists(file_path):
                general_utils.save_object_as_json(data, file_path)
            else:
                general_utils.update_json_file(data, file_path)

            self._store_file_in_file_store(self.filestore_buckets[bucket_name], file_name, file_path)
        except Exception as e:
            print("Failed to store json %s in the file store. Reason: %s" % (str(file_name), str(e)))
            print("Stored %s in %s" % (str(file_name), str(self.temporary_storage_path)))

    def _store_spark_object_into_files(self, spark_object, file_name, bucket_name='objects'):
        try:
            if self.file_store_handler.empty():
                raise SkipToLocalException("No file storage set up in 'config.py'")

            file_path = os.path.join(self.temporary_storage_path, file_name)
            spark_object.save(file_path)

            general_utils.compress_folder_to_zip(file_path)
            file_path += ".zip"
            file_name += ".zip"

            self._store_file_in_file_store(self.filestore_buckets[bucket_name], file_name, file_path)
            os.remove(file_path)
        except SkipToLocalException:
            # storing file in local storage
            file_path = os.path.join(self.local_storage_path, file_name)
            spark_object.save(file_path)
            print("Stored spark model in file %s in local storage %s" % (str(file_name), str(self.local_storage_path)))
        except Exception as e:
            print("Failed to store spark model in file storage. Reason: " + str(e))
            print("Spark model is saved in %s storage path instead." % str(self.temporary_storage_path))
            self.temporary_storage_files.add(str(file_name))

    ##########################################################################
    #                                                                        #
    #                          Helper Functions                              #
    #                                                                        #
    ##########################################################################

    def set_output_folders(self):
        """
        This function set output folders. if file storage is exists, temporary output is created
        else output folder is created.
        """
        if not os.path.exists(self.local_storage_path):
            os.makedirs(self.local_storage_path)
        if not self.file_store_handler.empty() and not os.path.exists(self.temporary_storage_path):
            os.makedirs(self.temporary_storage_path)

    def clean_output_folders(self):
        """
        This function set output folders. if file storage is exists, temporary output is created
        else output folder is created.
        """
        try:
            if os.path.exists(self.local_storage_path) and not os.listdir(self.local_storage_path):
                os.rmdir(self.local_storage_path)
            if os.path.exists(self.temporary_storage_path) and not os.listdir(self.temporary_storage_path):
                os.rmdir(self.temporary_storage_path)
        except:
            pass

    def import_object_from_module(self, module, class_name):
        s = "from " + str(module) + " import " + class_name
        try:
            return s
        except Exception as e:
            print(e)

    def update_config(self, config_model=None, config_data=None):
        if config_data is not None:
            config_data_settings = config_data.get('data_settings', {})
            updated_data_settings = general_utils.recursive_dict_update(config_data_settings,
                                                                        self.config.get('data_settings', {}))
            self.config['data_settings'] = updated_data_settings

        if config_model is not None:
            config_model_model_settings = config_model.get('model_settings', {})
            updated_model_settings = general_utils.recursive_dict_update(config_model_model_settings,
                                                                         self.config.get('model_settings', {}))
            self.config['model_settings'] = updated_model_settings

    def validate_config(self):
        pipeline = self.job_settings.get('pipeline', None)
        if pipeline:
            asset_name = self.identity['asset_name']
            schema_file = "assets/{}/configs/{}_schema.json".format(asset_name, pipeline)
        else:
            raise ConfigKeyError('missing required field - "pipeline" in `task_settings`')
        if os.path.exists(schema_file):
            with open(schema_file, 'r') as f:
                schema = json.loads(f.read())
            try:
                jsonschema.validate(instance=self.config, schema=schema)
            except jsonschema.exceptions.ValidationError as ve:
                raise ConfigError('Error validating your config: {' + str(ve.message) + '}\n')

    def temporary_files_message(self):
        if not self.file_store_handler.empty() and len(self.temporary_storage_files) > 0:
            print("The Following files:", sep=" ")
            print(*self.temporary_storage_files, sep=", ")
            print("failed to saved on file storage and stored in %s instead." % str(self.temporary_storage_path))

    def log_to_file(self, run_id):
        logger_dir = self.local_storage_path if self.file_store_handler.empty() else self.temporary_storage_path
        logger_file_path = os.path.join(logger_dir, self.get_logger_filename(run_id))

        # create log handler
        log = logging.getLogger(self.job_id)
        log.setLevel(logging.DEBUG)
        handler = logging.FileHandler(logger_file_path)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(logging.Formatter('%(message)s'))
        log.addHandler(handler)
        sys.stdout = Logger(log, logging.DEBUG, self.job_id)

    def pipeline_start_print(self, run_id=None):
        self.start_time = time.strftime('%Y-%m-%d %H:%M:%S')
        self.last_time = self.start_time
        print('-------------------------------------- Starting analysis ----------------------------------------------')
        if run_id is None:
            print('Analysis `{}` began running at {}'.format(self.identity['asset_name'], self.start_time))
        else:
            print('Analysis `{}` began running at `{}`, with Run Id: `{}`'
                  .format(self.identity['asset_name'], self.start_time, run_id))
        print('-------------------------------------------------------------------------------------------------------')

    def pipeline_end_print(self, run_id=None):
        time_now = time.strftime('%Y-%m-%d %H:%M:%S')
        start_time_delta = datetime.datetime.strptime(
            time_now, '%Y-%m-%d %H:%M:%S') - datetime.datetime.strptime(self.start_time, '%Y-%m-%d %H:%M:%S')

        self.last_time = time_now
        print('------------------------------------ Analysis Summary Print -------------------------------------------')
        print('Job id: {}'.format(self.job_id))
        print('Asset Name: {}'.format(self.identity['asset_name']))
        print('Run id: {}'.format(run_id))
        # print('Local Config path (in case `run.py` was used): ', self.config_path)
        images = self.output_manager.get_images_files()
        if images:
            print('Images saved:  ')
            for img_name in images:
                print(' >> ', img_name)
        print('Current date and time: {}'.format(time_now))
        print('Analysis started running at: {}'.format(self.start_time))
        print('Elapsed Time from the beginning: {}'.format(str(start_time_delta)))
        print('-------------------------------------------------------------------------------------------------------')
        print('Job has completed successfully.')



from azureml.core import Workspace
from azureml.core.authentication import InteractiveLoginAuthentication


def init_workspace(tenant_id, subscription_id, resource_group, workspace_name):
    if tenant_id is not None:
        interactive_auth = InteractiveLoginAuthentication(tenant_id=tenant_id)
        workspace = Workspace(
            subscription_id=subscription_id,
            resource_group=resource_group,
            workspace_name=workspace_name,
            auth=interactive_auth)
    else:
        workspace = Workspace(
            subscription_id=subscription_id,
            resource_group=resource_group,
            workspace_name=workspace_name)
    return workspace



import copy
import datetime
import collections
import json
from abc import ABCMeta, abstractmethod
import pandas as pd
import numpy as np
from mlapp.config import settings
from mlapp.utils.automl import AutoMLResults
from mlapp.managers.pipeline_manager import pipeline
from mlapp.managers.io_manager import IOManager
from mlapp.utils.exceptions.framework_exceptions import UnsupportedFileType, MissingConnectionException
from mlapp.handlers.wrappers.database_wrapper import database_instance


class _UserManager(object):
    prediction_types = {
        'TRAIN': 1,
        'TEST': 2,
        'FORECAST': 3
    }

    def __init__(self, config, _input: IOManager, _output: IOManager, run_id: str, *args, **kwargs):
        self.local_storage_path = settings.get("local_storage_path", "output")
        self.run_id = run_id
        self.data_settings = config.get('data_settings', {})
        self.model_settings = config.get('model_settings', {})
        self.job_settings = config.get('job_settings', {})
        self.flow_settings = config.get('flow_settings', {})
        self._input_manager = _input
        self._output_manager = _output

    __metaclass__ = ABCMeta

    @abstractmethod
    def _get_manager_type(self):
        pass

    def _get_all_metadata(self):
        """
        returns all key value pairs saved to metadata.
        :return: dictionary
        """
        return self._input_manager.get_metadata()

    def _get_all_objects(self):
        """
        Returns all saved objects.
        :return: dictionary of objects.
        """
        return self._input_manager.get_objects()

    def save_metadata(self, key, value):
        """
        Saves a metadata value to storage.
        :param key: String.
        :param value: metadata includes all ""json serializable"" objects
        (i.e string, int, dictionaries, list and tuples)
        :return: None
        """
        self._output_manager.set_analysis_metadata_value(self._get_manager_type(), key, copy.deepcopy(value))

    def get_metadata(self, key, default_value=None):
        """
        returns metadata value given a key.
        :param key: String.
        :param default_value: any object, default is None.
        :return: metadata value.
        """
        metadata = self._input_manager.get_metadata_value(self._get_manager_type(), default_value)
        return metadata.get(key, default_value)

    def save_object(self, obj_name, obj_value, obj_type='pkl'):
        """
        Saves objects to storage.
        :param obj_name: String.
        :param obj_value: Obj.
        :param obj_type: String. One of MLApp supported file types: 'pkl', 'pyspark', 'tensorflow', 'keras', 'pytorch'.
        Default='pkl'.
        :return: None
        """
        supported_types = ['pkl', 'pyspark', 'tensorflow', 'keras', 'pytorch']

        if obj_type not in supported_types:
            raise UnsupportedFileType('Unsupported file type: ' + obj_type)
        obj = collections.defaultdict(dict)
        obj[self._get_manager_type()][obj_name] = obj_value     # no deep copy as obj value can be non python
        self._output_manager.add_objects(obj_type, dict(obj))

    def get_object(self, obj_name, default_value=None):
        """
        Returns an object given a key.
        :param obj_name: String.
        :param default_value: String.
        :return: Object
        """
        return self._input_manager.get_objects_value(self._get_manager_type(), {}).get(obj_name, default_value)

    def save_images(self, images):
        """
        Saves images to storage.
        :param images: dictionary of matplotlib/pyplot figures. Keys will be used as image names.
        :return: None.
        """
        self._output_manager.add_images(images)

    def save_image(self, image_name, image):
        """
        Saves image to storage.
        :param image_name: String.
        :param image: matplotlib/pyplot figure.
        :return: None
        """
        self._output_manager.add_image(image_name, image)

    def get_dataframe(self, key):
        """
        Returns the Data Frame by key name
        :return: df
        """
        return self._input_manager.get_dataframe(key)

    def save_dataframe(self, key, value, to_table=None):
        """
        Save a data frame to storage.
        :param key: String
        :param value: a DataFrame.
        :param to_table: String. Database table name. default: None.
        :return: None
        """
        self._output_manager.add_dataframe(key, value, to_table)

    def save_automl_result(self, results: AutoMLResults, obj_type='pkl'):
        """
        Saves an AutoMLResults object.
        :param results: AutoMLResults object
        :param obj_type: type of framework, Default: 'pkl'.
        Supports: 'pkl', 'pyspark', 'tensorflow', 'keras', 'pytorch'.
        :return: None
        """
        self.save_object('model', results.get_best_model(), obj_type=obj_type)
        self.save_images(results.get_figures())
        metadata = results.get_metadata()
        for key in results.get_metadata():
            self.save_metadata(key, metadata[key])

    def get_automl_result(self) -> AutoMLResults:
        """
        Returns an AutoMLResult object that was saved in a previous run
        :return: AutoMLResult object
        """
        result = AutoMLResults()
        result.best_model = self.get_object('model')
        result.best_model_metrics = self.get_metadata('scores', {})
        result.best_estimator = self.get_metadata('estimator_family')
        result.best_model_key = self.get_metadata('model_class_name')
        result.selected_features = self.get_metadata('selected_features_names', [])
        result.intercept = self.get_metadata('intercept')
        result.coefficients = self.get_metadata('coefficients', [])
        result.feature_selection = self.get_metadata('feature_selection')
        result.best_cv_score = self.get_metadata('cv_score')
        return result


class DataManager(_UserManager):
    def _get_manager_type(self):
        return 'data'

    @abstractmethod
    def load_train_data(self, *args):
        """
        Write your own logic to load your train data.
        :return: data object to be passed to next pipeline step
        """
        pass

    @abstractmethod
    def load_forecast_data(self, *args):
        """
        Write your own logic to load your forecast data.
        :return: data object to be passed to next pipeline step
        """
        pass

    @abstractmethod
    def clean_train_data(self, data):
        """
        Write your own logic to clean your train data.
        :param data: data object received from previous pipeline step
        :return: data object to be passed to next pipeline step
        """
        return data

    @abstractmethod
    def clean_forecast_data(self, data):
        """
        Write your own logic to clean your forecast data.
        :param data: data object received from previous pipeline step
        :return: data object to be passed to next pipeline step
        """
        return data

    @abstractmethod
    def transform_train_data(self, data):
        """
        Write your own logic to transform your train data.
        :param data: data object received from previous pipeline step
        :return: data object to be passed to next pipeline step
        """
        return data

    @abstractmethod
    def transform_forecast_data(self, data):
        """
        Write your own logic to transform your forecast data.
        :param data: data object received from previous pipeline step
        :return: data object to be passed to next pipeline step
        """
        return data

    @abstractmethod
    def load_target_data(self, *args):
        """
        Write your own logic to load your historical target actual values (for monitoring purposes).
        """
        pass

    @pipeline
    def cache_features(self, df):
        """
        Used to save the features dataframe. It can later be retrieved using get_features().
        :param df: Dataframe
        :return: None
        """
        self._output_manager.add_dataframe('features', df)

    @pipeline
    def load_features(self):
        """
        Returns the dataframe "features" from input manager.
        :return: df
        """
        return self._input_manager.get_dataframe('features')

    @pipeline
    def update_actuals(self, df):
        if database_instance.empty():
            raise MissingConnectionException('Failed to update actuals. No database configured in environment.')
        database_instance.update_actuals(df)

    def get_input_from_predecessor_job(self, key=None):
        """
        Returns output from predecessor job "features" from input manager (in a flow settings).
        :param key: key name of value
        :return: any object
        """
        if key is not None:
            return self._input_manager.metadata.get('input_from_predecessor', {}).get(key)
        else:
            return self._input_manager.metadata.get('input_from_predecessor', {})


class ModelManager(_UserManager):
    def _get_manager_type(self):
        return 'models'

    @abstractmethod
    def train_model(self, data):
        """
        Write your own logic to train your model.
        :param data: data object received from previous pipeline step
        :return: None
        """
        pass

    #########################################
    #    Should be implemented by child     #
    #########################################
    @abstractmethod
    def forecast(self, data):
        """
        Write your own logic for forecast.
        :param data: data object received from previous pipeline step
        :return: None
        """
        pass

    #########################################
    #    Should be implemented by child     #
    #########################################
    @abstractmethod
    def refit(self, data):
        """
        Write your own logic for refit.
        :param data: data object received from previous pipeline step
        :return: None
        """
        pass

    def add_predictions(self, primary_keys_columns, y_hat, y=pd.Series(dtype='float64'), prediction_type='TRAIN'):
        """
        Creates a prediction dataframe and saves it to storage.
        :param primary_keys_columns: array-like. shared index for y_hat and y_true.
        :param y_hat: Series / array-like. predictions.
        :param y: Series / array-like. True values.
        :param prediction_type: String. (i.e TRAIN, TEST, FORECAST)
        :return: None.
        """
        y_true = None
        prediction_type = prediction_type.upper()
        if y is not None and len(y)>0:
            y_true = y.copy()

        # Check if y_hat has more than 1 dimension:
        if (y_hat is not None) and (len(y_hat.shape) > 1):
            y_hat = y_hat.reshape(1, len(y_hat))[0]

        # If y_true is a dataframe with only one column, use that column:
        if isinstance(y_true, pd.core.frame.DataFrame):
            if len(y_true.columns.to_list()):
                y_true = y_true[y_true.columns[0]]
        try:
            prediction_type_int = self.prediction_types[prediction_type]
        except:
            prediction_type_int = len(self.prediction_types)+1

        new_prediction = pd.DataFrame({
            "y_true": y_true,
            "y_hat": y_hat.copy(),
            "type": np.repeat(prediction_type_int, len(y_hat)),
            "index": primary_keys_columns.copy()
        })

        self._output_manager.add_dataframe('predictions', new_prediction, 'target')

    @pipeline
    def evaluate_prediction_accuracy(self, *args):
        # TODO: add timestamp filter below to query and implement in postgres handler
        data = database_instance.get_model_predictions(model_id=self.job_settings['model_id'])

        from_timestamp = self.model_settings.get('from_timestamp', min(data['timestamp']))
        to_timestamp = self.model_settings.get('to_timestamp', max(data['timestamp']))

        filter_target_df = data[(data['timestamp'] >= from_timestamp) & (data['timestamp'] <= to_timestamp)]

        acc = eval(self.model_settings['model_accuracy_summary'])(
            filter_target_df['y_hat'], filter_target_df['y_true'], filter_target_df['y_hat'],
            filter_target_df['y_true'])

        # we don't need training and testing we just keep one, and remove the name from the key:
        acc = {
            k.split(' ')[1]: acc[k] for k in acc.keys()
            if k.startswith(np.unique([k.split(' ')[0] for k in acc.keys()])[0])
        }

        df = pd.DataFrame({
            'model_id': self.job_settings['model_id'],
            'asset_name': self.job_settings['asset_name'],
            'asset_label_name': self.job_settings.get('asset_label_name'),
            'created_at': datetime.datetime.now(),
            'updated_at': datetime.datetime.now(),
            'timestamp': datetime.datetime.now(),
            'model_accuracy': json.dumps(acc)
        }, index=[0])

        database_instance.insert_df('asset_accuracy_monitoring', df)




import pandas as pd
import numpy as np
from sklearn.metrics import accuracy_score, roc_auc_score, precision_score, recall_score, \
    fbeta_score, r2_score, mean_absolute_error, mean_squared_error


def regression(train_y_predict, test_y_predict, train_y_real, test_y_real, *args, **kwargs):
    """
    Return a dictionary of accuracy scores for provided predicted values.
    The following scores are returned: Training Accuracy(R^2), Testing Accuracy(R^2), Training MAPE, Testing MAPE.
    :param train_y_predict: y train predicted values.
    :param train_y_real: y train actual values.
    :param test_y_predict: y test predicted values.
    :param test_y_real: y test actual values.
    :return: dictionary with accuracy scores.
    """
    try:
        train_y_predict_ravel = np.ravel(train_y_predict)
        test_y_predict_ravel = np.ravel(test_y_predict)
        train_y_actuals_ravel = np.ravel(train_y_real)
        test_y_actuals_ravel = np.ravel(test_y_real)
    except Exception as err:
        raise ("Cannot run `np.ravel` on one of the inputs: " + str(err))

    return {
        'R2 (train set)': r2_score(train_y_actuals_ravel, train_y_predict_ravel),
        'R2 (test set)': r2_score(test_y_actuals_ravel, test_y_predict_ravel),
        'MAPE (train set)': _get_mean_absolute_percentage_error(train_y_predict_ravel, train_y_actuals_ravel),
        'MAPE (test set)': _get_mean_absolute_percentage_error(test_y_predict_ravel, test_y_actuals_ravel),
        'MAE (train set)': mean_absolute_error(train_y_actuals_ravel, train_y_predict_ravel),
        'MAE (test set)': mean_absolute_error(test_y_actuals_ravel, test_y_predict_ravel),
        'RMSE (train set)': np.sqrt(mean_squared_error(train_y_predict_ravel, train_y_actuals_ravel)),
        'RMSE (test set)': np.sqrt(mean_squared_error(test_y_predict_ravel, test_y_actuals_ravel))
    }


def classification(train_y_predict, test_y_predict, train_y_real, test_y_real, beta=1, unbalanced=False, binary=None,
                   *args, **kwargs):
    """
    Return a dictionary of accuracy scores for provided predicted values. The following scores are returned:
    Jaccard Score Training, Jaccard Score Testing, AUC Score Training, AUC Score Testing, F_Beta Training Accuracy,
    F_Beta Testing Accuracy, recall Training Accuracy, recall Testing Accuracy, precision Training Accuracy,
    precision Testing Accuracy
    :param train_y_predict: y train predicted values.
    :param train_y_real: y train actual values.
    :param test_y_predict: y test predicted values.
    :param test_y_real: y test actual values.
    :param beta: F beta parameter. If beta=1, f1-score will be calculated
    :param unbalanced: Boolean. For now, set unbalanced=False --> to be fixed.
    :param binary: Boolean. If binary=True, average='binary' for F-beta calculation, else average='weighted'.
    Please refer to https://scikit-learn.org/stable/modules/generated/sklearn.metrics.fbeta_score.html to understand the
    meaning
    :return: dictionary with accuracy scores.
    """
    if kwargs.get('test_y_proba') is not None and kwargs.get('train_y_proba') is not None:
        try:
            training_auc = float(roc_auc_score(train_y_real, kwargs['train_y_proba']))
            testing_auc = float(roc_auc_score(test_y_real, kwargs['test_y_proba']))
        except:
            print('INFO: scores summary no AUC accuracy')
            training_auc = None
            testing_auc = None
    else:
        training_auc = None
        testing_auc = None
    try:
        training_accuracy = float(accuracy_score(train_y_real, train_y_predict))
        testing_accuracy = float(accuracy_score(test_y_real, test_y_predict))
    except:
        print('INFO: Scores summary no accuracy')
        training_accuracy = None
        testing_accuracy = None

    if unbalanced:
        try:
            fbeta_score_train = _unbalanced_data(train_y_real, train_y_predict, beta=beta,
                                                 return_value=['precision', 'recall', 'F_beta'])
            fbeta_score_test = _unbalanced_data(test_y_real, test_y_predict, beta=beta,
                                                return_value=['precision', 'recall', 'F_beta'])
        except Exception as e:
            print(e)
            print('No F1-scoring for unbalanced')
            fbeta_score_train = None
            fbeta_score_test = None
    else:
        try:
            if binary is True:
                average = 'binary'
            elif binary is False:
                average = 'weighted'
            else:
                average = 'weighted' if len(pd.unique(train_y_real)) > 2 else 'binary'

            fbeta_score_train = {'recall': recall_score(train_y_real, train_y_predict, average=average),
                                 'precision': precision_score(train_y_real, train_y_predict, average=average),
                                 'F_beta': fbeta_score(train_y_real, train_y_predict, beta=beta, average=average)}
            fbeta_score_test = {'recall': recall_score(test_y_real, test_y_predict, average=average),
                                'precision': precision_score(test_y_real, test_y_predict, average=average),
                                'F_beta': fbeta_score(test_y_real, test_y_predict, beta=beta, average=average)}
        except Exception as e:
            print(e)
            print('No F1-scoring for unbalanced')
            fbeta_score_train = None
            fbeta_score_test = None

    scores = {}
    if training_accuracy is not None and testing_accuracy is not None:
        scores['jaccard (train set)'] = training_accuracy
        scores['jaccard (test set)'] = testing_accuracy
    if training_auc is not None and testing_auc is not None:
        scores['AUC (train set)'] = training_auc
        scores['AUC (test set)'] = testing_auc
    if fbeta_score_train is not None and fbeta_score_test is not None:
        scores['f1_score (train set)'] = fbeta_score_train.get('F_beta')
        scores['f1_score (test set)'] = fbeta_score_test.get('F_beta')
        scores['recall (train set)'] = fbeta_score_train.get('recall')
        scores['recall (test set)'] = fbeta_score_test.get('recall')
        scores['precision (train set)'] = fbeta_score_train.get('precision')
        scores['precision (test set)'] = fbeta_score_test.get('precision')

    return scores


def time_series(train_y_predict, test_y_predict, train_y_real, test_y_real, *args, **kwargs):
    """
    Return a dictionary of accuracy scores for provided predicted values, specifically in case of time series challenges.
    The following scores are returned: Training Accuracy(R^2), Testing Accuracy(R^2), Training MAPE, Testing MAPE,
    Training MASE, Testing MASE.
    :param train_y_predict: y train predicted values.
    :param train_y_real: y train actual values.
    :param test_y_predict: y test predicted values.
    :param test_y_real: y test actual values.
    :return: dictionary with accuracy scores.
    """
    return {'R2 (train set)': r2_score(train_y_real, train_y_predict),
            'R2 (test set)': r2_score(test_y_real, test_y_predict),
            'MAPE (train set)': _get_mean_absolute_percentage_error(train_y_predict, train_y_real),
            'MAPE (test set)': _get_mean_absolute_percentage_error(test_y_predict, test_y_real),
            'MASE (train set)': _get_mean_average_scaled_error(train_y_predict, train_y_real),
            'MASE (test set)': _get_mean_average_scaled_error(test_y_predict, test_y_real)}


def _get_mean_absolute_percentage_error(y_predictions, y_actuals):
    '''
    Mean absolute percentage error (MAPE) calculations
    :param y_predictions: labels estimations
    :param y_actuals: true labels
    :return: MAPE (float)
    '''
    y_actuals[np.where(y_actuals == 0)] = 1
    return np.mean(np.abs((y_actuals - y_predictions) / y_actuals)) * 100


def _get_mean_average_scaled_error(y_predictions, y_actuals, peak_dates=None):
    '''
    Mean average scaled error (MASE) calculations
    :param y_predictions: labels estimations
    :param y_actuals: true labels
    :param peak_dates: list of peak dates to exclude from calculations
    :return: MASE (float)
    '''
    y_a = y_actuals.copy()
    y_p = y_predictions.copy()
    if peak_dates is None:
        naive_model_error = np.mean(np.abs(y_a[1:] - np.roll(y_a, 1)[1:]))
    else:
        # remove zero index if exists because can't "naive" predict it
        peak_dates_no_zero_index = peak_dates.copy()
        zero_index = np.argwhere(peak_dates == 0)
        if len(zero_index) > 0:
            peak_dates_no_zero_index = np.delete(peak_dates_no_zero_index, zero_index)
        naive_model_error = np.mean(
            np.abs(y_actuals[peak_dates_no_zero_index] - y_actuals[peak_dates_no_zero_index - 1]))
        y_a = y_actuals[peak_dates_no_zero_index]
        y_p = y_predictions[peak_dates_no_zero_index]

    # MASE
    if naive_model_error != 0:
        mase = np.mean(np.abs(y_a - y_p) / naive_model_error)
    # MAD/Mean Ratio (when naive_model_error is zero)
    else:
        y_actuals_mean = np.mean(y_a)
        if y_actuals_mean == 0:
            y_actuals_mean = y_actuals_mean + 1
        mase = np.mean(np.abs(y_a - y_p) / y_actuals_mean)
    return mase


def _unbalanced_data(y_true, y_pred, beta=1, negative_label=None, return_value=None):
    """
    Returns scores in case of unbalanced data (all categories in target columns are not represented the same).
    This function is meant to overweight low-represented categories (or anomalies) in order to emphasize the ability of
    the model to detect specific cluster.
    :param y_true: true labels
    :param y_pred: estimated labels
    :param beta: beta of the F-beta calculatation. By default, beta=1 (F1-score)
    :param negative_label: List of labels to unit for the negative labels. The rest will be considered as positive.
    :param return_value: List of scores to return: 'F_beta', 'precision', or 'recall'
    :return: Dictionary with all the scores.
    """
    if negative_label is None:
        negative_label = [0]
    if return_value is None:
        return_value = ['F_beta']

    if isinstance(y_true, pd.DataFrame):
        y_true_reshaped = y_true.iloc[:, 0]
    elif not isinstance(y_true, pd.Series):
        y_true_reshaped = pd.Series(y_true)
    else:
        y_true_reshaped = y_true.copy()

    if isinstance(y_pred, pd.DataFrame):
        y_pred_reshaped = y_pred.iloc[:, 0]
    elif not isinstance(y_pred, pd.Series):
        y_pred_reshaped = pd.Series(y_pred, index=y_true.index)
    else:
        y_pred_reshaped = y_pred.copy()

    # Get the positive and the negative labels
    positive_labels = list(set(y_true_reshaped.values) - set(negative_label))

    # Replace negative_labels by 0 and positive labels by 1
    y_true_reshaped = y_true_reshaped.replace(negative_label, 0)
    y_pred_reshaped = y_pred_reshaped.replace(negative_label, 0)
    y_true_reshaped = y_true_reshaped.replace(positive_labels, 1)
    y_pred_reshaped = y_pred_reshaped.replace(positive_labels, 1)

    # Split the positive predictions and the negative predictions
    positive_predictions = y_true_reshaped[y_true_reshaped != 0] - y_pred_reshaped[y_true_reshaped != 0]
    negative_predictions = y_true_reshaped[y_true_reshaped == 0] - y_pred_reshaped[y_true_reshaped == 0]

    # tp/tn/fp/fn
    TP = len(positive_predictions[positive_predictions == 0])
    TN = len(negative_predictions[negative_predictions == 0])
    FP = len(negative_predictions[negative_predictions != 0])
    FN = len(positive_predictions[positive_predictions != 0])

    # recall and precision
    outputs = {
        'recall': float(TP) / (TP + FN) if (TP + FN) != 0 else 1.0,
        'precision': float(TP) / (TP + FP) if (TP + FP) != 0 else 1.0
    }

    # F beta score
    if outputs['precision'] == 0 and outputs['recall'] == 0:
        outputs['F_beta'] = 0
    else:
        outputs['F_beta'] = (1 + np.square(beta)) * outputs['precision'] * outputs['recall'] / \
                            (np.square(beta) * outputs['precision'] + outputs['recall'])
    if isinstance(return_value, str):
        if return_value not in outputs:
            raise Exception("Unknown return value: " + return_value)
        return outputs[return_value]

    elif isinstance(return_value, list):
        return_output = {}
        for value in return_value:
            if value not in outputs:
                raise Exception("Unknown value in 'return_value' list!")
            else:
                return_output[value] = outputs[value]
        return return_output

    else:
        raise Exception("Unknown type for 'return_value'!")


from mlapp.mlapp_cli.common.cli_utilities import create_directory, str_to_camelcase, str_capitalize
from mlapp.mlapp_cli.common.files import data_manager_file, model_manager_file, train_config_file, \
    forecast_config_file, train_config_file_with_flow, forecast_config_file_with_flow
import os, click, shutil, uuid, re
import numpy as np
from mlapp.mlapp_cli.common.model_rename_dictionary import rename_dictionary
from mlapp.mlapp_cli.common.cli_utilities import create_file, copy_files
from mlapp.mlapp_cli.cli_help import cli_assets_help


@click.group("assets")
def commands():
    """
    ML App Assets Command

    Use it to install or create your Assets.
    type --help on each assets commands to get more information.

    """
    pass


@commands.command("show", help=cli_assets_help['show'])
def show():
    assets_folder_path = os.path.join(os.getcwd(), 'assets')
    if os.path.exists(assets_folder_path):
        your_assets = list(filter(lambda asset_name: os.path.isdir(os.path.join(assets_folder_path, asset_name)),
                                  os.listdir(assets_folder_path)))
        if len(your_assets) == 0:
            click.secho(
                "Your project not contains any assets at the moment.\nHint: use mlapp assets 'mlapp boilerplates install BOILERPLATE_NAME' or 'mlapp assets create ASSET_NAME' command for generating your first asset." ,fg='red')
        else:
            click.echo('Your project assets:')
            for asset in your_assets:
                if os.path.isdir(os.path.join(assets_folder_path, asset)) and asset != '__pycache__':
                    click.echo('[*] ' + asset)
    else:
        click.secho(
            "ERROR: your project is not empty.\nHint: you can use 'mlapp init --force' option to force init (caution: force may override exsiting files).", fg='red')


@commands.command("create", help=cli_assets_help['create'])
@click.argument("name", required=True)
@click.option("-f", "--force", is_flag=True, default=False,
              help="Flag force will override existing asset_name file.")
@click.option("-w", "--with-flow", is_flag=True, default=False,
              help="Creates asset train and forecast configs with flow settings.")
def create(name, force, with_flow):
    # model name validations and transformations.
    asset_name = name.lower()
    asset_name = asset_name.replace(" ", "_")
    asset_name = asset_name.replace("-", "_")

    # creates the assets directory if not exists.
    create_directory(directory_name='assets')

    full_directory_path = os.path.join(os.getcwd(), os.path.join('assets', asset_name))
    if os.path.exists(full_directory_path) and not force:
        click.secho(
            'Error: ' + asset_name + ' asset already exists.\nHint: please select a unique name to your asset or use --force option to override asset folder.', fg='red')
        return

    # create the necessary folders for the new asset, `asset_name` directory and asset configs directory.
    create_directory(directory_name=asset_name, path='assets')
    create_directory(directory_name='configs', path=os.path.join('assets', asset_name), include_init=False)

    model_name_capitalized = str_to_camelcase(asset_name
                                              )
    # generates model manager file template
    model_manager_file_content = model_manager_file.format(model_name_capitalized)

    # generates data manager file template
    data_manager_file_content = data_manager_file.format(model_name_capitalized)

    if with_flow:
        # generates train config file template with flow
        train_config_file_content = train_config_file_with_flow.replace("<ASSET_NAME>", asset_name)

        # generates forecast config file template with flow
        forecast_config_file_content = forecast_config_file_with_flow.replace("<ASSET_NAME>", asset_name)
    else:
        # generates train config file template
        train_config_file_content = train_config_file.replace("<ASSET_NAME>", asset_name)

        # generates forecast config file template
        forecast_config_file_content = forecast_config_file.replace("<ASSET_NAME>", asset_name)

    # create all managers templates
    create_file(file_name=asset_name + '_model_manager.py', path=os.path.join('assets', asset_name), permissions='w+',
                content=model_manager_file_content)
    create_file(file_name=asset_name + '_data_manager.py', path=os.path.join('assets', asset_name), permissions='w+',
                content=data_manager_file_content)

    # create all configs templates
    create_file(file_name=asset_name + '_train_config.json',
                path=os.path.join(os.path.join('assets/', asset_name), 'configs'),
                permissions='w+',
                content=train_config_file_content)
    create_file(file_name=asset_name + '_forecast_config.json',
                path=os.path.join(os.path.join('assets/', asset_name), 'configs'),
                permissions='w+',
                content=forecast_config_file_content)


@commands.command("rename", help=cli_assets_help['rename'])
@click.argument("name", required=True)
@click.argument("new_name", required=True)
@click.option('-d', '--delete', is_flag=True, default=False,
              help="Use it to delete previous asset directory on renaming.")
def rename(name, new_name, dir_name=None, show_success_msg=True, delete=False, force=False):
    if name is None or new_name is None:
        click.secho(
            "ERROR: asset old name and new name are required.\nHint: 'mlapp assets rename OLD_NAME NEW_NAME' is the right command format.", fg='red')
        return True

    if name == new_name:
        click.secho("ERROR: asset old name and new name are equal. please use a different name.", fg='red')
        return True

    except_list = []
    if dir_name is None:
        old_name_path = os.path.join(os.getcwd(), 'assets', name)
    else:
        old_name_path = os.path.join(os.getcwd(), 'assets', dir_name)
    new_name_path = os.path.join(os.getcwd(), 'assets', new_name)

    if not os.path.exists(old_name_path):
        click.secho(
            f"ERROR: asset \"{name}\" is not exists.\nHint: use 'mlapp assets show' command to view all available assets.", fg='red')
    elif os.path.exists(new_name_path) and not force:
        click.secho(
            f"ERROR: asset \"{new_name}\" is already exists.", fg='red')
    else:

        # creates the new model directory
        os.mkdir(new_name_path)

        # base renaming
        if name in rename_dictionary.keys():
            model_dict = rename_dictionary.get(name, {})
            replace_files = model_dict.keys()
        else:
            model_dict = rename_dictionary.get('base', {})
            replace_files = model_dict.keys()

        # rename model name file by file.
        for rf in replace_files:
            try:
                file_additional_path = model_dict[rf].get('inner_path', '')

                except_list.append(name + rf)
                # files path
                if dir_name is None:
                    full_old_filename = os.path.join(os.getcwd(), "assets", name, file_additional_path,
                                                     name + rf)
                else:
                    full_old_filename = os.path.join(os.getcwd(), "assets", dir_name, file_additional_path,
                                                     name + rf)
                full_new_filename = os.path.join(os.getcwd(), "assets", new_name, file_additional_path, new_name + rf)

                # checks if needs to create sub directories.
                if not os.path.exists(os.path.join(os.getcwd(), "assets", new_name, file_additional_path)):
                    os.mkdir(os.path.join(os.getcwd(), "assets", new_name, file_additional_path))

                words_to_replace = model_dict[rf].get('words', [])

                # read file content
                file_content = ''''''
                with open(full_old_filename, 'r') as file:
                    lines = file.readlines()
                    for line in lines:
                        # line_striped = line.strip()

                        # replace new file path
                        for wr_d in words_to_replace:
                            word = wr_d.get('word', None)
                            word_type = wr_d.get('word_type', 'append-left')
                            word_format = wr_d.get('word_format', 'str.lower')
                            word_pattern = wr_d.get('word_pattern', None)
                            if word is not None:
                                # check word format and type
                                word_format_func = eval(word_format)
                                old_name_formatted = word_format_func(name)
                                new_name_formatted = word_format_func(new_name)

                                if word_type == 'append-left':
                                    if word_pattern is not None and isinstance(word_pattern, str):
                                        word_pattern = old_name_formatted + word_pattern
                                    elif word_pattern is not None and (
                                            isinstance(word_pattern, list) or isinstance(word_pattern, np.array)):
                                        wp_res = ''
                                        for wp in word_pattern:
                                            wp_res += old_name_formatted + wp + '|'
                                        word_pattern = wp_res[:-1]

                                    old_name_formatted = old_name_formatted + word
                                    new_name_formatted = new_name_formatted + word
                                else:
                                    # 'append-right'

                                    # check for patterns type
                                    if word_pattern is not None and isinstance(word_pattern, str):
                                        word_pattern += old_name_formatted
                                    elif word_pattern is not None and (
                                            isinstance(word_pattern, list) or isinstance(word_pattern, np.array)):
                                        wp_res = ''
                                        for wp in word_pattern:
                                            wp_res += wp + old_name_formatted + '|'
                                        word_pattern = wp_res[:-1]

                                    old_name_formatted = word + old_name_formatted
                                    new_name_formatted = word + new_name_formatted

                                # checking and replacing model names or using regex
                                if word_pattern is not None:
                                    line = re.sub(word_pattern, new_name_formatted, line)
                                elif old_name_formatted in line:
                                    line = line.replace(old_name_formatted, new_name_formatted)

                        file_content += line

                # creates the new file with the new name.
                create_file(file_name=full_new_filename, content=file_content)

            except Exception as e:
                shutil.rmtree(new_name_path)
                click.echo(e)
                # click.echo('ERROR: Oops something bad happened.')
                return True

        # copy the rest of the files
        copy_files(old_name_path, new_name_path, except_list)

        if delete:
            shutil.rmtree(old_name_path)

        if show_success_msg:
            click.secho('Success: asset "' + name + '" was renamed to "' + new_name + '".', fg='green')